# =============================================================================
# Minet Twitter Tweet Search CLI Action
# =============================================================================
#
# Logic of the `tw tweet-search` action.
#
from twitwi import normalize_tweets_payload_v2, format_tweet_as_csv_row
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.twitter.utils import with_twitter_client
from minet.cli.twitter.utils import validate_query_boundaries

ITEMS_PER_PAGE = 100


@with_enricher_and_loading_bar(
    headers=TWEET_FIELDS,
    title="Retrieving tweets",
    unit="queries",
    nested=True,
    sub_unit="tweets",
)
@with_twitter_client(api_version="2")
def action(cli_args, client, enricher, loading_bar):
    validate_query_boundaries(cli_args)

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(query):
            kwargs = {
                "query": query,
                "max_results": ITEMS_PER_PAGE,
                "sort_order": cli_args.sort_order,
                "expansions": ",".join(TWEET_EXPANSIONS),
                "params": TWEET_PARAMS,
            }

            if cli_args.start_time:
                kwargs["start_time"] = cli_args.start_time
            if cli_args.end_time:
                kwargs["end_time"] = cli_args.end_time
            if cli_args.since_id:
                kwargs["since_id"] = cli_args.since_id
            if cli_args.until_id:
                kwargs["until_id"] = cli_args.until_id

            route = (
                ["tweets", "search", "all"]
                if cli_args.academic
                else ["tweets", "search", "recent"]
            )

            while True:
                try:
                    result = client.call(route, **kwargs)
                except TwitterHTTPError as e:
                    if e.e.code == 404:
                        enricher.writerow(row)
                    else:
                        raise e

                    continue

                # Empty response
                if (
                    result["meta"]["result_count"] == 0
                    and "next_token" in result["meta"]
                ):
                    kwargs["next_token"] = result["meta"]["next_token"]
                    continue

                normalized_tweets = normalize_tweets_payload_v2(
                    result, collection_source="api"
                )

                for normalized_tweet in normalized_tweets:
                    addendum = format_tweet_as_csv_row(normalized_tweet)
                    enricher.writerow(row, addendum)
                    loading_bar.nested_advance()

                if "next_token" in result["meta"]:
                    kwargs["next_token"] = result["meta"]["next_token"]
                else:
                    break



from minet.cli.argparse import command

# TODO: lazyloading issue
from minet.constants import DEFAULT_THROTTLE

CRAWL_COMMAND = command(
    "crawl",
    "minet.cli.crawl.crawl",
    title="Minet Crawl Command",
    description="""
        Use multiple threads to crawl the web using minet crawling and
        scraping DSL.
    """,
    epilog="""
        Examples:

        . Running a crawler definition:
            $ minet crawl crawler.yml -O crawl-data
    """,
    no_output=True,
    arguments=[
        {"name": "crawler", "help": "Path to the crawler definition file."},
        {
            "flags": ["-O", "--output-dir"],
            "help": "Output directory.",
            "default": "crawl",
        },
        {
            "flag": "--resume",
            "help": "Whether to resume an interrupted crawl.",
            "action": "store_true",
        },
        {
            "flag": "--dump-queue",
            "help": "Print the contents of the persistent queue. (This is for debug only, don't use this flag unless you know what you are doing).",
            "action": "store_true",
        },
        {
            "flag": "--throttle",
            "help": "Time to wait - in seconds - between 2 calls to the same domain. Defaults to %s."
            % DEFAULT_THROTTLE,
            "type": float,
            "default": DEFAULT_THROTTLE,
        },
    ],
)



from minet.wikipedia import WikimediaRestAPIClient
from minet.cli.utils import with_enricher_and_loading_bar, with_ctrl_c_warning


def get_headers(cli_args):
    if cli_args.sum:
        return ["views"]

    return ["timestamp", "views"]


@with_enricher_and_loading_bar(
    headers=get_headers, title="Collecting pageviews", unit="pages"
)
@with_ctrl_c_warning
def action(cli_args, enricher, loading_bar):
    client = WikimediaRestAPIClient()

    lang_pos = (
        enricher.headers[cli_args.lang_column]
        if cli_args.lang_column is not None
        else None
    )

    page_pos = enricher.headers[cli_args.column]

    def key(row):
        if lang_pos is not None:
            lang = row[lang_pos]
        else:
            lang = cli_args.lang

        return (lang, row[page_pos])

    for row, pageviews in client.pageviews(
        enricher.rows(),
        key=key,
        start_date=cli_args.start_date,
        end_date=cli_args.end_date,
        granularity=cli_args.granularity,
        access=cli_args.access,
        agent=cli_args.agent,
        threads=cli_args.threads,
    ):
        with loading_bar.step():

            if cli_args.sum:
                s = 0

                for item in pageviews:
                    s += item.views

                enricher.writerow(row, [s])
                continue

            for item in pageviews:
                enricher.writerow(row, [item.timestamp, item.views])



# =============================================================================
# Minet Web Utilities
# =============================================================================
#
# Miscellaneous web-related functions used throughout the library.
#
from typing import Optional, Callable, Tuple, Union, OrderedDict, List, Any, Dict

import re
import cgi
import certifi
import random
import browser_cookie3
import urllib3
import urllib3.exceptions as urllib3_exceptions
import urllib.error
import ural
import json
import mimetypes
import functools
import warnings
from bs4 import BeautifulSoup, XMLParsedAsHTMLWarning, SoupStrainer
from datetime import datetime
from timeit import default_timer as timer
from io import BytesIO
from threading import Event
from itertools import chain
from urllib.parse import urljoin
from urllib3 import HTTPResponse
from urllib3.util.ssl_ import create_urllib3_context
from urllib.request import Request
from ebbe import rcompose, noop
from collections import defaultdict
from tenacity import (
    Retrying,
    RetryCallState,
    retry_if_exception_type,
    retry_if_exception,
    stop_after_attempt,
    stop_when_event_set,
    sleep_using_event,
)
from tenacity.wait import wait_base

from minet.encodings import normalize_encoding
from minet.loggers import sleepers_logger
from minet.utils import is_binary_mimetype, infer_encoding
from minet.exceptions import (
    RedirectError,
    MaxRedirectsError,
    InfiniteRedirectsError,
    InvalidRedirectError,
    InvalidURLError,
    SelfRedirectError,
    CancelledRequestError,
    FinalTimeoutError,
)
from minet.constants import (
    DEFAULT_SPOOFED_UA,
    DEFAULT_SPOOFED_TLS_CIPHERS,
    DEFAULT_URLLIB3_TIMEOUT,
    COOKIE_BROWSERS,
)

mimetypes.init()

# Types
AnyTimeout = Union[float, urllib3.Timeout]

# Handy regexes
CHARSET_RE = re.compile(rb'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
PRAGMA_RE = re.compile(rb'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
XML_RE = re.compile(rb'^\s*<\?xml.*?encoding=["\']*(.+?)["\'>]', flags=re.I)
NOSCRIPT_RE = re.compile(rb"<noscript[^>]*>.*</noscript[^>]*>", flags=re.I)
META_REFRESH_RE = re.compile(
    rb"""<meta\s+http-equiv=['"]?refresh['"]?\s+content=['"]?([^"']+)['">]?""",
    flags=re.I,
)
JAVASCRIPT_LOCATION_RE = re.compile(
    rb"""(?:window\.)?location(?:\s*=\s*|\.replace\(\s*)['"`](.*?)['"`]"""
)
ESCAPED_SLASH_RE = re.compile(rb"\\\/")
HTML_RE = re.compile(
    rb"^\s*<(?:html|head|body|title|meta|link|span|div|img|ul|ol|[ap!?])", flags=re.I
)
CANONICAL_LINK_RE = re.compile(
    rb'<link\s*[^>]*\s+rel=(?:"\s*canonical\s*"|canonical|\'\s*canonical\s*\')\s+[^>]*\s?/?>'
)
HREF_RE = re.compile(rb'href=(\"[^"]+|\'[^\']+|[^\s]+)>?\s?', flags=re.I)

# Constants
REDIRECT_STATUSES = set(HTTPResponse.REDIRECT_STATUSES)
CONTENT_PREBUFFER_UP_TO = 1024
STREAMING_CHUNK_SIZE: int = 2**12
LARGE_CONTENT_PREBUFFER_UP_TO = 2**16
EXPECTED_WEB_ERRORS = (
    urllib3_exceptions.HTTPError,
    RedirectError,
    InvalidURLError,
    FinalTimeoutError,
)

assert CONTENT_PREBUFFER_UP_TO < LARGE_CONTENT_PREBUFFER_UP_TO


def infer_encoding_from_headers(response: urllib3.HTTPResponse) -> Optional[str]:
    content_type_header = response.getheader("content-type")

    if content_type_header is None:
        return None

    parsed_header = cgi.parse_header(content_type_header)

    if len(parsed_header) > 1:
        charset = parsed_header[1].get("charset")

        if charset is not None:
            return charset.lower()

    return None


def infer_encodings_from_xml(
    body: bytes, chunk_size: Optional[int] = CONTENT_PREBUFFER_UP_TO
) -> Dict[str, int]:
    possibilities = defaultdict(list)

    chunk = body

    if chunk_size is not None:
        chunk = body[:chunk_size]

    matches = chain(
        re.findall(CHARSET_RE, chunk),
        re.findall(PRAGMA_RE, chunk),
        re.findall(XML_RE, chunk),
    )

    for match in matches:
        encoding = match.decode().lower()
        possibilities[normalize_encoding(encoding)].append(encoding)

    return {max(names, key=len): len(names) for names in possibilities.values()}


def looks_like_html(html_chunk: bytes) -> bool:
    return HTML_RE.match(html_chunk) is not None


def parse_http_header(header: str) -> Tuple[str, str]:
    key, value = header.split(":", 1)

    return key.strip(), value.strip()


# TODO: take more cases into account...
#   http://www.otsukare.info/2015/03/26/refresh-http-header
def parse_http_refresh(value) -> Optional[Tuple[int, str]]:
    try:

        if isinstance(value, bytes):
            value = value.decode()

        duration, url = value.strip().split(";", 1)

        if not url.lower().strip().startswith("url="):
            return None

        return int(duration), str(url.split("=", 1)[1])
    except Exception:
        return None


def extract_href(value):

    m = HREF_RE.search(value)

    if not m:
        return None

    url = m.group(1)

    try:
        url = url.decode("utf-8")
    except UnicodeDecodeError:
        return None

    return url.strip("\"'") or None


def find_canonical_link(html_chunk: bytes):
    m = CANONICAL_LINK_RE.search(html_chunk)

    if not m:
        return None

    return extract_href(m.group())


def find_meta_refresh(html_chunk: bytes):
    m = META_REFRESH_RE.search(html_chunk)

    if not m:
        return None

    return parse_http_refresh(m.group(1))


def find_javascript_relocation(html_chunk: bytes):
    m = JAVASCRIPT_LOCATION_RE.search(html_chunk)

    if not m:
        return None

    try:
        return ESCAPED_SLASH_RE.sub(b"/", m.group(1)).decode()
    except Exception:
        return None


class CookieResolver(object):
    def __init__(self, jar):
        self.jar = jar

    def __call__(self, url):
        req = Request(url)
        self.jar.add_cookie_header(req)

        return req.get_header("Cookie") or None


def grab_cookies(browser="firefox"):
    if browser not in COOKIE_BROWSERS:
        raise TypeError('minet.utils.grab_cookies: unknown "%s" browser.' % browser)

    try:
        return CookieResolver(getattr(browser_cookie3, browser)())
    except browser_cookie3.BrowserCookieError as e:
        raise e


def dict_to_cookie_string(d):
    return "; ".join("%s=%s" % r for r in d.items())


def create_pool_manager(
    proxy: Optional[str] = None,
    threads: Optional[int] = None,
    insecure: bool = False,
    spoof_tls_ciphers: bool = False,
    **kwargs
) -> urllib3.PoolManager:
    """
    Helper function returning a urllib3 pool manager with sane defaults.
    """

    manager_kwargs: Dict[str, Any] = {"timeout": DEFAULT_URLLIB3_TIMEOUT}

    if not insecure:
        manager_kwargs["cert_reqs"] = "CERT_REQUIRED"
        manager_kwargs["ca_certs"] = certifi.where()
    else:
        manager_kwargs["cert_reqs"] = "CERT_NONE"
        # manager_kwargs['assert_hostname'] = False

        urllib3.disable_warnings()

    if threads is not None:

        # TODO: maxsize should increase with group_parallelism
        manager_kwargs["maxsize"] = 10
        manager_kwargs["num_pools"] = threads * 2

    manager_kwargs.update(kwargs)

    if spoof_tls_ciphers:
        manager_kwargs["ssl_context"] = create_urllib3_context(
            ciphers=DEFAULT_SPOOFED_TLS_CIPHERS
        )

    if proxy is not None:
        return urllib3.ProxyManager(proxy, **manager_kwargs)

    return urllib3.PoolManager(**manager_kwargs)


DEFAULT_POOL_MANAGER = create_pool_manager(maxsize=10, num_pools=10)


def stream_request_body(
    response: urllib3.HTTPResponse,
    body: BytesIO,
    chunk_size: int = STREAMING_CHUNK_SIZE,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
    up_to: Optional[int] = None,
) -> bool:
    if cancel_event is not None and cancel_event.is_set():
        raise CancelledRequestError

    if up_to is not None and body.tell() >= up_to:
        return False

    for data in response.stream(chunk_size):
        body.write(data)

        if up_to is not None and body.tell() >= up_to:
            return False

        if cancel_event is not None and cancel_event.is_set():
            raise CancelledRequestError

        if final_time is not None:
            if timer() >= final_time:
                raise FinalTimeoutError

    # This is the only place we know the body has been fully read
    return True


def timeout_to_final_time(timeout: AnyTimeout) -> float:
    seconds: float

    if isinstance(timeout, urllib3.Timeout):
        if timeout.total is not None:
            seconds = timeout.total
        else:
            seconds = timeout.connect_timeout + timeout.read_timeout
    else:
        seconds = timeout

    # Some epsilon so sockets can timeout themselves properly
    seconds += 0.01

    return timer() + seconds


def pool_manager_aware_timeout_to_final_time(
    pool_manager: urllib3.PoolManager, timeout: Optional[AnyTimeout]
) -> Optional[float]:
    if timeout is not None:
        return timeout_to_final_time(timeout)

    else:
        pool_manager_timeout = pool_manager.connection_pool_kw.get("timeout")

        if pool_manager_timeout is not None:
            return timeout_to_final_time(pool_manager_timeout)

    return None


class BufferedResponse(object):
    """
    Class wrapping a urllib3.HTTPResponse and representing a response whose
    body has not been yet fully read.

    It is a low-level representation used in the minet.web module that should
    not leak outside normally.

    It is able to stream a response's body safely all while keeping track of
    a "final" timeout correctly enforced to bypass python socket race condition
    issues on read loops and is also able to be cancelled if required.
    """

    __slots__ = (
        "__inner",
        "__body",
        "__cancel_event",
        "__final_time",
        "__finished",
        "__unwrapped",
    )

    def __init__(
        self,
        response: urllib3.HTTPResponse,
        cancel_event: Optional[Event],
        final_time: Optional[float] = None,
    ):
        self.__inner = response
        self.__cancel_event = cancel_event
        self.__final_time = final_time
        self.__body = BytesIO()
        self.__finished = False
        self.__unwrapped = False

    def __len__(self) -> int:
        return self.__body.getbuffer().nbytes

    def __stream(
        self, chunk_size: int = STREAMING_CHUNK_SIZE, up_to: Optional[int] = None
    ) -> None:
        if self.__unwrapped:
            raise TypeError("buffered response was already unwrapped")

        if self.__finished:
            return

        fully_read = False

        try:
            fully_read = stream_request_body(
                self.__inner,
                chunk_size=chunk_size,
                cancel_event=self.__cancel_event,
                final_time=self.__final_time,
                body=self.__body,
                up_to=up_to,
            )
        finally:
            if fully_read:
                self.__finished = fully_read
                self.close()

    def geturl(self) -> Optional[str]:
        return self.__inner.geturl()

    def getheader(self, name: str, default: Optional[str] = None) -> Optional[str]:
        return self.__inner.getheader(name, default)

    @property
    def status(self) -> int:
        return self.__inner.status

    @property
    def body(self) -> bytes:
        return self.__body.getvalue()

    def close(self) -> None:
        # NOTE: releasing and closing is a noop if already done
        self.__inner.release_conn()
        self.__inner.close()

    def unwrap(self) -> Tuple[urllib3.HTTPResponse, bytes]:
        self.__unwrapped = True
        self.close()
        return self.__inner, self.__body.getvalue()

    def read(self, chunk_size: int = STREAMING_CHUNK_SIZE) -> None:
        self.__stream(chunk_size=chunk_size)

    def prebuffer_up_to(
        self, amount: int, chunk_size: int = STREAMING_CHUNK_SIZE
    ) -> None:
        self.__stream(chunk_size=chunk_size, up_to=amount)


def atomic_request(
    pool_manager: urllib3.PoolManager,
    url: str,
    method="GET",
    headers=None,
    timeout=None,
    body=None,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
) -> BufferedResponse:
    """
    Generic request helpers using a urllib3 pool_manager to access some resource.
    """

    # Validating URL
    if not ural.is_url(
        url, require_protocol=True, tld_aware=True, allow_spaces_in_path=True
    ):
        raise InvalidURLError(url=url)

    # Performing request
    request_kwargs = {
        "headers": headers,
        "body": body,
        "preload_content": False,
        "release_conn": False,
        "redirect": False,
        "retries": False,
    }

    if timeout is not None:
        request_kwargs["timeout"] = timeout

    if final_time is None:
        final_time = pool_manager_aware_timeout_to_final_time(pool_manager, timeout)

    response = pool_manager.request(method, url, **request_kwargs)

    return BufferedResponse(response, cancel_event=cancel_event, final_time=final_time)


class Redirection(object):
    __slots__ = ("status", "type", "url")

    def __init__(self, url: str, _type="hit"):
        self.status: Optional[int] = None
        self.url: str = url
        self.type: str = _type

    def __repr__(self):
        class_name = self.__class__.__name__

        return ("<%(class_name)s type=%(type)s status=%(status)s url=%(url)s>") % {
            "class_name": class_name,
            "type": self.type,
            "status": self.status,
            "url": self.url,
        }


RedirectionStack = List[Redirection]


def atomic_resolve(
    pool_manager: urllib3.PoolManager,
    url: str,
    method: str = "GET",
    headers=None,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    infer_redirection: bool = False,
    timeout: Optional[AnyTimeout] = None,
    cancel_event: Optional[Event] = None,
    final_time: Optional[float] = None,
    body=None,
    canonicalize: bool = False,
) -> Tuple[RedirectionStack, BufferedResponse]:
    """
    Helper function attempting to resolve the given url.
    """

    url_stack: OrderedDict[str, Redirection] = OrderedDict()
    error: Optional[Exception] = None
    buffered_response: Optional[BufferedResponse] = None

    if final_time is None:
        final_time = pool_manager_aware_timeout_to_final_time(pool_manager, timeout)

    for _ in range(max_redirects):
        if infer_redirection:
            target = ural.infer_redirection(url, recursive=False)

            if target != url:
                url_stack[url] = Redirection(url, "infer")
                url = target
                continue

        # We close last buffered_response as it won't be used anymore
        if buffered_response:
            buffered_response.close()

        # We check for cancellation
        if cancel_event is not None and cancel_event.is_set():
            raise CancelledRequestError

        redirection = Redirection(url)

        try:
            buffered_response = atomic_request(
                pool_manager,
                url,
                method=method,
                headers=headers,
                body=body,
                timeout=timeout,
                final_time=final_time,
                cancel_event=cancel_event,
            )

        # Request error
        except urllib3_exceptions.HTTPError as e:
            url_stack[url] = redirection
            error = e
            break

        # Cycle
        if url in url_stack:
            error = InfiniteRedirectsError("Infinite redirects")
            break

        redirection.status = buffered_response.status
        url_stack[url] = redirection

        # Attempting to find next location
        location = None

        if buffered_response.status not in REDIRECT_STATUSES:

            if buffered_response.status < 400:

                # Refresh header
                if follow_refresh_header:
                    refresh = buffered_response.getheader("refresh")

                    if refresh is not None:
                        p = parse_http_refresh(refresh)

                        if p is not None:
                            location = p[1]
                            redirection.type = "refresh-header"

                # Reading a small chunk of the html
                if location is None and (follow_meta_refresh or follow_js_relocation):
                    try:
                        buffered_response.prebuffer_up_to(CONTENT_PREBUFFER_UP_TO)
                    except Exception as e:
                        if isinstance(e, CancelledRequestError):
                            raise
                        error = e
                        break

                # Meta refresh
                if location is None and follow_meta_refresh:
                    meta_refresh = find_meta_refresh(buffered_response.body)

                    if meta_refresh is not None:
                        location = meta_refresh[1]
                        redirection.type = "meta-refresh"

                # JavaScript relocation
                if location is None and follow_js_relocation:
                    js_relocation = find_javascript_relocation(buffered_response.body)

                    if js_relocation is not None:
                        location = js_relocation
                        redirection.type = "js-relocation"

            # Found the end
            if location is None:
                redirection.type = "hit"

            # Canonical url
            if redirection.type == "hit":
                if canonicalize:
                    try:
                        buffered_response.prebuffer_up_to(LARGE_CONTENT_PREBUFFER_UP_TO)
                    except Exception as e:
                        if isinstance(e, CancelledRequestError):
                            raise
                        error = e
                        break

                    canonical = find_canonical_link(buffered_response.body)

                    if canonical is not None and canonical != url:
                        canonical = urljoin(url, canonical)
                        redirection = Redirection(canonical, "canonical")
                        url_stack[canonical] = redirection

                break

        else:
            redirection.type = "location-header"
            location = buffered_response.getheader("location")

        # Invalid redirection
        if not location:
            error = InvalidRedirectError("Redirection is invalid")
            break

        # Location badly encoded?
        try:
            if not location.isascii():
                byte_location = location.encode("latin1")
                encoding = infer_encoding(byte_location)

                if (
                    encoding is not None
                    and encoding != "iso-8859-1"
                    and encoding != "latin1"
                    and encoding != "ascii"
                ):
                    location = byte_location.decode(encoding)
        except Exception:
            pass

        # Resolving next url
        next_url = urljoin(url, location.strip())

        # Self loop?
        if next_url == url:
            error = SelfRedirectError("Self redirection")
            break

        # Go to next
        url = next_url

    # We reached max redirects
    else:
        error = MaxRedirectsError("Maximum number of redirects exceeded")

    # NOTE: error is raised that late to be sure we cleanup resources attached
    # to the connection
    if error is not None:
        if buffered_response is not None:
            buffered_response.close()
        raise error

    compiled_stack = list(url_stack.values())

    assert buffered_response is not None

    return compiled_stack, buffered_response


def build_request_headers(headers=None, cookie=None, spoof_ua=False, json_body=False):

    # Formatting headers
    final_headers = {}

    if spoof_ua:
        final_headers["User-Agent"] = DEFAULT_SPOOFED_UA

    if cookie:
        if not isinstance(cookie, str):
            cookie = dict_to_cookie_string(cookie)

        final_headers["Cookie"] = cookie

    if json_body:
        final_headers["Content-Type"] = "application/json"

    # Note: headers passed explicitly by users always win
    if headers is not None:
        final_headers.update(headers)

    return final_headers


class Response(object):
    """
    Class representing a finalized HTTP response.

    It wraps a urllib3.HTTPResponse as well as its raw body and exposes a
    variety of useful utilities that will be used downstream.

    This class is used by high-level function of this module and is expected
    to be found elsewhere.

    Note that it will lazily compute required items when asked for certain
    properties.

    It also works as a kind of dict that can bring around meta information.
    """

    __slots__ = (
        "__response",
        "__stack",
        "__body",
        "__text",
        "__url",
        "__meta",
        "__datetime_utc",
        "__is_text",
        "__encoding",
        "__ext",
        "__mimetype",
        "__has_guessed_extension",
        "__has_guessed_encoding",
        "__has_decoded_text",
    )

    __response: urllib3.HTTPResponse
    __stack: Optional[RedirectionStack]
    __body: bytes
    __text: Optional[str]
    __url: str
    __meta: Dict[str, Any]
    __datetime_utc: datetime
    __is_text: Optional[bool]
    __encoding: Optional[str]
    __ext: Optional[str]
    __mimetype: Optional[str]

    __has_guessed_extension: bool
    __has_guessed_encoding: bool
    __has_decoded_text: bool

    def __init__(
        self,
        url: str,
        stack: Optional[RedirectionStack],
        response: urllib3.HTTPResponse,
        body: bytes,
        known_encoding: Optional[str] = None,
    ):
        self.__url = url
        self.__stack = stack
        self.__response = response
        self.__body = body
        self.__text = None
        self.__meta = {}
        self.__datetime_utc = datetime.utcnow()
        self.__is_text = None
        self.__encoding = known_encoding
        self.__ext = None
        self.__mimetype = None

        self.__has_guessed_extension = False
        self.__has_guessed_encoding = known_encoding is not None
        self.__has_decoded_text = False

    def __guess_extension(self) -> None:
        if self.__has_guessed_extension:
            return

        # Guessing mime type
        # TODO: validate mime type string?
        url = self.__url
        assert url is not None
        mimetype, _ = mimetypes.guess_type(url)

        response = self.__response

        if "Content-Type" in response.headers:
            content_type = response.headers["Content-Type"]
            parsed_header = cgi.parse_header(content_type)

            if parsed_header and parsed_header[0].strip():
                mimetype = parsed_header[0].strip()

                # Dropping charset info
                if "," in mimetype:
                    mimetype = mimetype.split(",", 1)[0]

        if mimetype is None and looks_like_html(self.__body):
            mimetype = "text/html"

        if mimetype is not None:
            ext = mimetypes.guess_extension(mimetype)

            if ext == ".htm":
                ext = ".html"
            elif ext == ".jpe":
                ext = ".jpg"

            self.__mimetype = mimetype
            self.__ext = ext

        if mimetype is not None:
            self.__is_text = not is_binary_mimetype(mimetype)

        self.__has_guessed_extension = True

    def __guess_encoding(self) -> None:
        if self.__has_guessed_encoding:
            return

        self.__guess_extension()

        if not self.__is_text:
            return

        self.__encoding = infer_encoding(self.__body)

        self.__has_guessed_encoding = True

    def __decode(self, errors: str = "replace") -> None:
        if self.__has_decoded_text:
            return

        self.__guess_extension()
        self.__guess_encoding()

        assert self.__is_text is not None

        if not self.__is_text:
            raise TypeError("response is binary and cannot be decoded")

        self.__text = self.__body.decode(self.__encoding or "utf-8", errors=errors)
        self.__has_decoded_text = True

    @property
    def url(self) -> str:
        return self.__url

    @property
    def start_url(self) -> str:
        return self.__url

    @property
    def end_url(self) -> str:
        if self.__stack is None:
            return self.__url
        return self.__stack[-1].url

    @property
    def headers(self):
        return self.__response.headers

    @property
    def stack(self) -> Optional[RedirectionStack]:
        return self.__stack

    @property
    def was_redirected(self) -> bool:
        return self.start_url != self.end_url

    @property
    def status(self) -> int:
        return self.__response.status

    @property
    def end_datetime(self) -> datetime:
        return self.__datetime_utc

    @property
    def ext(self) -> Optional[str]:
        self.__guess_extension()
        return self.__ext

    @property
    def mimetype(self) -> Optional[str]:
        self.__guess_extension()
        return self.__mimetype

    @property
    def is_text(self) -> bool:
        self.__guess_extension()
        return self.__is_text  # type: ignore # At that point we know it's a bool

    @property
    def could_be_html(self) -> bool:
        mime = self.mimetype
        return mime is not None and "/htm" in mime

    @property
    def is_html(self) -> bool:
        return self.is_text and looks_like_html(self.__body)

    @property
    def encoding(self) -> Optional[str]:
        self.__guess_encoding()
        return self.__encoding

    @property
    def likely_encoding(self) -> str:
        encoding = self.encoding

        if encoding is None:
            return "utf-8"

        return encoding

    @property
    def encoding_from_headers(self) -> Optional[str]:
        return infer_encoding_from_headers(self.__response)

    # TODO: add encoding_from_xml & possible_encodings when required

    @property
    def body(self) -> bytes:
        return self.__body

    def text(self) -> str:
        self.__decode()
        return self.__text  # type: ignore # If we don't raise, this IS a string

    def json(self):
        return json.loads(self.text())

    def soup(
        self,
        engine: str = "lxml",
        ignore_xhtml_warning=False,
        strainer: Optional[SoupStrainer] = None,
    ) -> BeautifulSoup:
        with warnings.catch_warnings():
            if ignore_xhtml_warning:
                warnings.simplefilter("ignore", category=XMLParsedAsHTMLWarning)

            return BeautifulSoup(self.text(), engine, parse_only=strainer)

    def __getitem__(self, name: str) -> Any:
        return self.__meta[name]

    def __setitem__(self, name: str, value: Any) -> None:
        self.__meta[name] = value

    def __contains__(self, name: str) -> bool:
        return name in self.__meta

    def get(self, name: str, default=None) -> Optional[Any]:
        return self.__meta.get(name, default)

    def set(self, name: str, value: Any) -> None:
        self.__meta[name] = value


def request(
    url: str,
    pool_manager: urllib3.PoolManager = DEFAULT_POOL_MANAGER,
    method: str = "GET",
    headers=None,
    cookie=None,
    spoof_ua: bool = True,
    follow_redirects: bool = True,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    canonicalize: bool = False,
    known_encoding: Optional[str] = None,
    timeout: Optional[AnyTimeout] = None,
    body=None,
    json_body=None,
    cancel_event: Optional[Event] = None,
) -> Response:

    # Formatting headers
    final_headers = build_request_headers(
        headers=headers,
        cookie=cookie,
        spoof_ua=spoof_ua,
        json_body=json_body is not None,
    )

    # Dealing with body
    final_body = None

    if isinstance(body, bytes):
        final_body = body
    elif isinstance(body, str):
        final_body = body.encode("utf-8")

    if json_body is not None:
        final_body = json.dumps(json_body, ensure_ascii=False).encode("utf-8")

    stack: Optional[RedirectionStack] = None

    if not follow_redirects:
        buffered_response = atomic_request(
            pool_manager,
            url,
            method,
            headers=final_headers,
            body=final_body,
            timeout=timeout,
            cancel_event=cancel_event,
        )
    else:
        stack, buffered_response = atomic_resolve(
            pool_manager,
            url,
            method,
            headers=final_headers,
            body=final_body,
            max_redirects=max_redirects,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            canonicalize=canonicalize,
            timeout=timeout,
            cancel_event=cancel_event,
        )

    buffered_response.read()
    response, body = buffered_response.unwrap()

    return Response(url, stack, response, body, known_encoding=known_encoding)


def resolve(
    url: str,
    pool_manager: urllib3.PoolManager = DEFAULT_POOL_MANAGER,
    method: str = "GET",
    headers=None,
    cookie=None,
    spoof_ua: bool = True,
    max_redirects: int = 5,
    follow_refresh_header: bool = True,
    follow_meta_refresh: bool = False,
    follow_js_relocation: bool = False,
    infer_redirection: bool = False,
    timeout: Optional[AnyTimeout] = None,
    canonicalize: bool = False,
    cancel_event: Optional[Event] = None,
) -> RedirectionStack:

    final_headers = build_request_headers(
        headers=headers, cookie=cookie, spoof_ua=spoof_ua
    )

    stack, buffered_response = atomic_resolve(
        pool_manager,
        url,
        method,
        headers=final_headers,
        max_redirects=max_redirects,
        follow_refresh_header=follow_refresh_header,
        follow_meta_refresh=follow_meta_refresh,
        follow_js_relocation=follow_js_relocation,
        infer_redirection=infer_redirection,
        timeout=timeout,
        canonicalize=canonicalize,
        cancel_event=cancel_event,
    )

    buffered_response.close()

    return stack


def request_jsonrpc(
    url: str,
    method: str,
    pool_manager: urllib3.PoolManager = DEFAULT_POOL_MANAGER,
    *args,
    **kwargs
) -> Response:
    params = []

    if len(args) > 0:
        params = args
    elif len(kwargs) > 0:
        params = kwargs

    return request(
        url,
        pool_manager=pool_manager,
        method="POST",
        json_body={"method": method, "params": params},
        known_encoding="utf-8",
    )


ONE_DAY = 24 * 60 * 60


class log_request_retryer_before_sleep:
    def __init__(
        self, epilog_builder: Optional[Callable[[RetryCallState], str]] = None
    ):
        if epilog_builder is not None and not callable(epilog_builder):
            raise TypeError("epilog_builder should be None or callable")

        self.epilog_builder = epilog_builder

    def __call__(self, retry_state: RetryCallState) -> None:
        assert retry_state.outcome is not None

        if not retry_state.outcome.failed:
            raise NotImplementedError

        exception = retry_state.outcome.exception()

        sleepers_logger.warn(
            "request_retryer starts sleeping",
            extra={
                "source": "request_retryer",
                "retry_state": retry_state,
                "exception": exception,
                "sleep_time": retry_state.next_action.sleep,
                "epilog": self.epilog_builder(retry_state)
                if self.epilog_builder is not None
                else None,
            },
        )


class request_retryer_custom_exponential_backoff(wait_base):
    def __init__(self, min: float = 10, max: float = ONE_DAY, exp_base=4) -> None:
        self.min = min
        self.max = max
        self.exp_base = exp_base

    def __call__(self, retry_state: RetryCallState) -> float:
        # NOTE: we add/subtract randomly up to 1/4 of expected sleep time
        jitter = 0.5 - random.random()

        try:
            exp = self.exp_base ** (retry_state.attempt_number - 1)
            result = self.min * exp
            result += result / 2 * jitter
        except OverflowError:
            result = self.max

        return max(0, min(result, self.max))


def create_request_retryer(
    min: float = 10,
    max: float = ONE_DAY,
    max_attempts: int = 9,
    before_sleep=noop,
    additional_exceptions=None,
    retry_on_timeout: bool = True,
    predicate: Optional[Callable[[BaseException], bool]] = None,
    epilog: Optional[Callable[[RetryCallState], str]] = None,
    cancel_event: Optional[Event] = None,
):

    # By default we only retry network issues, such as Internet being cut off etc.
    retryable_exception_types = [
        urllib3_exceptions.ProtocolError,
        urllib.error.URLError,
    ]

    # We also usually include most timeout errors
    if retry_on_timeout:
        retryable_exception_types.extend(
            [FinalTimeoutError, urllib3_exceptions.TimeoutError]
        )

    if additional_exceptions:
        for exc in additional_exceptions:
            retryable_exception_types.append(exc)

    retry_condition = retry_if_exception_type(
        exception_types=tuple(retryable_exception_types)
    )

    if callable(predicate):
        retry_condition |= retry_if_exception(predicate)

    retrying_kwargs = {
        "reraise": True,
        "wait": request_retryer_custom_exponential_backoff(min=min, max=max),
        "retry": retry_condition,
        "stop": stop_after_attempt(max_attempts),
        "before_sleep": rcompose(
            log_request_retryer_before_sleep(epilog), before_sleep
        ),
    }

    # Cancellable sleep?
    if cancel_event is not None:
        if cancel_event.is_set():
            raise TypeError("cannot retry using an already set cancel_event")

        retrying_kwargs["sleep"] = sleep_using_event(cancel_event)
        retrying_kwargs["stop"] |= stop_when_event_set(cancel_event)
        retrying_kwargs["retry"] &= retry_if_exception(
            lambda _: not cancel_event.is_set()
        )

    return Retrying(**retrying_kwargs)


def retrying_method(attr="retryer"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            retryer = getattr(self, attr)

            if not isinstance(retryer, Retrying):
                raise ValueError

            return retryer(fn, self, *args, **kwargs)

        return decorated

    return decorate



# =============================================================================
# Minet Facebook Post Stats CLI Action
# =============================================================================
#
# Logic of the `fb post-stats` action.
#

# NOTE: this command is probably defunct, and even if it still works it needs
# to be ported to new loading bar scheme
import re
import json5
import casanova
from bs4 import BeautifulSoup
from datetime import datetime
from ural.facebook import is_facebook_post_url
from ebbe import getpath

from minet.utils import sleep_with_entropy
from minet.web import request
from minet.cli.utils import print_err, LoadingBar
from minet.cli.exceptions import FatalError
from minet.facebook.constants import (
    FACEBOOK_WEB_DEFAULT_THROTTLE,
    FACEBOOK_REACTION_KEYS,
)

META_EXTRACTOR_TEMPLATE = rb'bigPipe\.onPageletArrive\((\{.+share_fbid:"%s".+\})\);\}\)'
HTML_EXTRACTOR_TEMPLATE = rb"<code[^>]*><!--(.+%s.+)--></code>"
CURRENT_AVAILABILITY_DISCLAIMER = b"The link you followed may have expired, or the page may only be visible to an audience"
AVAILABILITY_DISCLAIMER = (
    b"The link you followed may be broken, or the page may have been removed"
)
LOGIN_DISCLAIMER = b"You must log in to continue"
CAPTCHA = b'id="captcha"'

REPORT_HEADERS = [
    "error",
    "canonical",
    "account_name",
    "timestamp",
    "time",
    "link",
    "aria_label",
    "text",
    "share_count",
    "comment_count",
    "reaction_count",
    "video_view_count",
]

for emotion_name in FACEBOOK_REACTION_KEYS.values():
    REPORT_HEADERS.append("%s_count" % emotion_name)

ERROR_PADDING = [""] * (len(REPORT_HEADERS) - 1)


def get_count(item):
    if item is None:
        return 0

    value = item.get("count")

    if not value:
        value = item.get("total_count")

    return value or 0


def collect_top_reactions(data):
    edges = getpath(data, ["top_reactions", "edges"])

    if edges is None:
        return

    index = {}

    for edge in edges:
        emotion = FACEBOOK_REACTION_KEYS.get(edge["node"]["key"])

        if emotion is None:
            print_err("Found unkown emotion %s" % edge)
            continue

        index[emotion] = edge["reaction_count"] or 0

    return index


def format_err(err):
    return [err] + ERROR_PADDING


def format(data):
    video_view_count = data.get("video_view_count")
    scraped = data["scraped"] or {}

    row = [
        "",
        data.get("url", ""),
        scraped.get("account_name", ""),
        scraped.get("timestamp", ""),
        scraped.get("time", ""),
        scraped.get("link", ""),
        scraped.get("aria_label", ""),
        scraped.get("text", ""),
        get_count(data["share_count"]),
        get_count(data["comment_count"]),
        get_count(data["reaction_count"]),
        video_view_count if isinstance(video_view_count, int) else "",
    ]

    emotion_index = collect_top_reactions(data)

    for emotion_name in FACEBOOK_REACTION_KEYS.values():
        row.append(emotion_index.get(emotion_name, 0))

    return row


def action(cli_args):
    enricher = casanova.enricher(
        cli_args.input, cli_args.output, add=REPORT_HEADERS, keep=cli_args.select
    )

    def fetch_facebook_page_stats(url):
        try:
            response = request(url, cookie="locale=en_US")
        except Exception:
            return "http-error", None

        if response.status == 404:
            return "not-found", None

        if response.status >= 400:
            return "http-error", None

        html = response.data

        if CAPTCHA in html:
            raise FatalError(["Rate limit reached!", "Last url: %s" % url])

        if CURRENT_AVAILABILITY_DISCLAIMER in html or AVAILABILITY_DISCLAIMER in html:
            return "unavailable", None

        if LOGIN_DISCLAIMER in html:
            return "private-or-unavailable", None

        # TODO: integrate into ural
        bpost_id = url.rsplit("/", 1)[-1].encode()

        # Extracting metadata
        meta_extractor = re.compile(META_EXTRACTOR_TEMPLATE % bpost_id)

        match = meta_extractor.search(html)

        if match is None:
            return "extraction-failed", None

        data = json5.loads(match.group(1).decode())
        data = getpath(
            data,
            [
                "jsmods",
                "pre_display_requires",
                0,
                3,
                1,
                "__bbox",
                "result",
                "data",
                "feedback",
            ],
        )

        if data is None:
            return "extraction-failed", None

        # TODO: remove, this is here as a test
        # TODO: try to find a post where comments are disabled
        if get_count(data["seen_by_count"]):
            print_err(
                "Found seen_by_count: %i for %s"
                % (get_count(data["seen_by_count"]), url)
            )

        if "political_figure_data" in data and data["political_figure_data"]:
            print_err("Found political_figure_data:")
            print_err(data["political_figure_data"])

        if get_count(data["reaction_count"]) != get_count(data["reactors"]):
            print_err("Found different reactions/reactors for %s" % url)

        # Extracting data from hidden html
        hidden_html_extractor = re.compile(HTML_EXTRACTOR_TEMPLATE % bpost_id)
        match = hidden_html_extractor.search(html)

        if match is not None:
            hidden_html = match.group(1).decode()
            soup = BeautifulSoup(hidden_html, "lxml")

            # Sometimes fetching a post behaves weirdly
            if soup.select_one("h5 a") is None:
                return "extraction-failed", None

            data["scraped"] = {}

            timestamp_elem = soup.select_one("[data-utime]")
            timestamp = int(timestamp_elem.get("data-utime"))

            data["scraped"]["account_name"] = soup.select_one("h5 a").get_text().strip()
            data["scraped"]["timestamp"] = timestamp
            data["scraped"]["time"] = datetime.fromtimestamp(timestamp).isoformat()

            # TODO: use a context manager
            try:
                data["scraped"]["aria_label"] = timestamp_elem.parent.get("aria-label")
            except Exception:
                pass

            try:
                data["scraped"]["text"] = soup.select_one(
                    '[data-testid="post_message"]'
                ).get_text()
            except Exception:
                pass

            # try:
            #     data['scraped']['link'] = soup.select_one('[data-lynx-uri]').get('href')
            # except Exception:
            #     pass

        return None, data

    # Loading bar
    loading_bar = LoadingBar(
        desc="Fetching post stats", unit="post", total=cli_args.total
    )

    for row, post_url in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.update()

        if not post_url or not is_facebook_post_url(post_url):
            enricher.writerow(row, format_err("not-facebook-post"))
            continue

        err, data = fetch_facebook_page_stats(post_url)

        if err:
            enricher.writerow(row, format_err(err))
        else:
            enricher.writerow(row, format(data))

        # Throttling
        sleep_with_entropy(FACEBOOK_WEB_DEFAULT_THROTTLE, 5.0)



# =============================================================================
# Minet Telegram Scraper
# =============================================================================
#
# Functions to scrape Telegram.
#
from bs4 import BeautifulSoup
import re

from minet.utils import (
    rate_limited_method,
    RateLimiterState,
    clean_human_readable_numbers,
)
from minet.web import (
    create_pool_manager,
    request,
    create_request_retryer,
    retrying_method,
)

from minet.telegram.constants import (
    TELEGRAM_URL,
    TELEGRAM_DEFAULT_THROTTLE,
)
from ural.telegram import (
    convert_telegram_url_to_public,
    parse_telegram_url,
    TelegramChannel as ParsedTelegramChannel,
    TelegramGroup as ParsedTelegramGroup,
    TelegramMessage as ParsedTelegramMessage,
)
from minet.telegram.formatters import TelegramChannelInfos, TelegramChannelMessages
from minet.telegram.exceptions import TelegramInvalidTargetError

TELEGRAM_IMG_RE = re.compile(r"background-image:url\(\'(.*?)\'\)")
TELEGRAM_HASHTAG_RE = re.compile(r"\?q=%23(.*)")


def forge_telegram_channel_url(name):
    url = TELEGRAM_URL + name + "/1"
    return url


def forge_telegram_channel_url_next(url, next_after):
    url = TELEGRAM_URL + next_after
    return url


def scrape_channel_infos(html):
    soup = BeautifulSoup(html, "lxml")

    channel_infos = soup.select_one("div[class='tgme_channel_info']")

    if not channel_infos:
        raise TelegramInvalidTargetError

    channel_header_infos = channel_infos.select_one(
        "div[class='tgme_channel_info_header']"
    )

    img = channel_header_infos.select_one("img").get("src")
    title = channel_header_infos.select_one("span").get_text()
    name = channel_header_infos.select_one(
        "div[class='tgme_channel_info_header_username']"
    ).get_text()
    link = (
        channel_header_infos.select_one(
            "div[class='tgme_channel_info_header_username']"
        )
        .select_one("a")
        .get("href")
    )

    channel_description = channel_infos.select_one(
        "div[class='tgme_channel_info_description']"
    )
    if channel_description:
        description = channel_description.get_text()

    # NOTE: the counter_type can be singular or plural, so we normalize
    counters_infos = {
        counter.select_one("span[class='counter_type']")
        .get_text()
        .rstrip("s"): counter.select_one("span[class='counter_value']")
        .get_text()
        for counter in channel_infos.select("div[class='tgme_channel_info_counter']")
    }

    nb_subscribers = counters_infos["subscriber"]
    nb_photos = counters_infos["photo"]
    nb_videos = counters_infos["video"]
    nb_links = counters_infos["link"]

    return TelegramChannelInfos(
        title=title,
        name=name,
        link=link,
        img=img,
        description=description,
        nb_subscribers=clean_human_readable_numbers(nb_subscribers),
        nb_photos=clean_human_readable_numbers(nb_photos),
        nb_videos=clean_human_readable_numbers(nb_videos),
        nb_links=clean_human_readable_numbers(nb_links),
    )


def scrape_channel_messages(html):
    results = []
    soup = BeautifulSoup(html, "lxml")

    messages_section = soup.select_one(
        "section[class='tgme_channel_history js-message_history']"
    )

    if not messages_section:
        raise TelegramInvalidTargetError

    next_after = messages_section.select_one("a[data-after]")
    if next_after:
        next_after = next_after.get("href")

    messages = messages_section.select(
        "div[class*='tgme_widget_message_wrap js-widget_message_wrap']"
    )
    for message in messages:

        if message.select_one(
            "div[class='tgme_widget_message text_not_supported_wrap service_message js-widget_message']"
        ):
            continue

        link_message = (
            TELEGRAM_URL
            + "s/"
            + (
                message.select_one(
                    "div[class*='tgme_widget_message text_not_supported_wrap']"
                ).get("data-post")
            )
        )

        user = message.select_one(
            "a[class='tgme_widget_message_owner_name']"
        ).get_text()
        user_link = (
            message.select_one("div[class='tgme_widget_message_user']")
            .select_one("a")
            .get("href")
        )
        user_img = (
            message.select_one("div[class='tgme_widget_message_user']")
            .select_one("img")
            .get("src")
        )

        links = []
        hashtags = []
        nb_links = 0
        nb_hashtags = 0
        text = ""
        texts = message.select("div[class*='tgme_widget_message_text js-message_text']")
        if texts:
            for text_html in texts:
                text += text_html.get_text()
                links_html = text_html.select("a")
                if links_html:
                    for link in links_html:
                        if TELEGRAM_HASHTAG_RE.match(link.get("href")):
                            hashtags.append(
                                TELEGRAM_HASHTAG_RE.match(link.get("href")).group(1)
                            )
                            nb_hashtags += 1
                        else:
                            links.append(link.get("href"))
                            nb_links += 1
        links = "|".join(links)
        hashtags = "|".join(hashtags)

        reply_img = None
        reply_user = None
        reply_text = None
        reply_html = message.select_one("a[class='tgme_widget_message_reply']")
        if reply_html:
            reply_img = reply_html.select_one(
                "i[class='tgme_widget_message_reply_thumb']"
            )
            if reply_img:
                reply_img = TELEGRAM_IMG_RE.search(reply_img.get("style")).group(1)
            reply_user = reply_html.select_one(
                "div[class='tgme_widget_message_author accent_color']"
            ).get_text()
            reply_text = reply_html.select_one(
                "div[class='tgme_widget_message_metatext js-message_reply_text']"
            )
            if reply_text:
                reply_text = reply_text.get_text()

        stickers = None
        stickers_html = message.select_one(
            "div[class='tgme_widget_message_sticker_wrap media_supported_cont']"
        )
        if stickers_html:
            stickers = stickers_html.select_one(
                "source[type='application/x-tgsticker']"
            )
            if stickers:
                stickers = stickers.get("srcset")
            else:
                stickers = stickers_html.select_one(
                    "i[class*='tgme_widget_message_sticker js-sticker_image']"
                )
                if stickers:
                    stickers = stickers.get("data-webp")

        nb_photos = 0
        photos = []
        photos_html = message.select(
            "a[class='tgme_widget_message_photo_wrap grouped_media_wrap blured js-message_photo']"
        )
        if photos_html:
            for photo in photos_html:
                photos.append(TELEGRAM_IMG_RE.search(photo.get("style")).group(1))
                nb_photos += 1
            photos = "|".join(photos)
        else:
            photos = message.select_one("a[class*='tgme_widget_message_photo_wrap']")
            if photos:
                photos = TELEGRAM_IMG_RE.search(photos.get("style")).group(1)
                nb_photos = 1

        nb_videos = 0
        videos_times = None
        videos = None
        video_html = message.select_one("a[class*='tgme_widget_message_video_player']")
        if video_html:
            video = video_html.select_one(
                "video[class='tgme_widget_message_video js-message_video']"
            )
            if video:
                videos = video.get("src")
            else:
                videos = "media is too big"
            video_time = video_html.select_one(
                "time[class='message_video_duration js-message_video_duration']"
            )
            if video_time:
                videos_times = video_time.get_text()
            nb_videos = 1

        link_img = None
        link_site = None
        link_title = None
        link_description = None
        link_html = message.select_one("a[class='tgme_widget_message_link_preview']")
        if link_html:
            link_img = link_html.select_one("i[class='link_preview_right_image']")
            if link_img:
                link_img = TELEGRAM_IMG_RE.search(link_img.get("style")).group(1)
            link_site = link_html.select_one(
                "div[class='link_preview_site_name accent_color']"
            )
            if link_site:
                link_site = link_site.get_text()
            link_title = link_html.select_one("div[class='link_preview_title']")
            if link_title:
                link_title = link_title.get_text()
            link_description = link_html.select_one(
                "div[class='link_preview_description']"
            )
            if link_description:
                link_description = link_description.get_text()

        views = message.select_one("span[class='tgme_widget_message_views']")
        could_be_displayed = bool(views)
        if views:
            views = views.get_text()
        datetime = (
            message.select_one("a[class='tgme_widget_message_date']")
            .select_one("time")
            .get("datetime")
        )
        edited_html = message.select_one(
            "span[class='tgme_widget_message_meta']"
        ).get_text()
        edited = "edited" in edited_html

        result = TelegramChannelMessages(
            link_to_message=link_message,
            could_be_displayed=could_be_displayed,
            user=user,
            user_link=user_link,
            user_img=user_img,
            text=text,
            nb_hashtags=nb_hashtags,
            hashtags=hashtags,
            is_reply_img=reply_img,
            is_reply_user=reply_user,
            is_reply_text=reply_text,
            stickers=stickers,
            nb_photos=nb_photos,
            photos=photos,
            nb_videos=nb_videos,
            videos=videos,
            videos_times=videos_times,
            nb_links=nb_links,
            links=links,
            link_img=link_img,
            link_site=link_site,
            link_title=link_title,
            link_description=link_description,
            views=clean_human_readable_numbers(views),
            datetime=datetime,
            edited=edited,
        )

        results.append(result)

    return next_after, results


class TelegramScraper(object):
    def __init__(self, throttle=TELEGRAM_DEFAULT_THROTTLE):
        self.pool_manager = create_pool_manager()

        self.rate_limiter_state = RateLimiterState(1, throttle)
        self.retryer = create_request_retryer()

    @rate_limited_method()
    @retrying_method()
    def request_page(self, url):
        response = request(url, pool_manager=self.pool_manager, known_encoding="utf-8")

        return response.text()

    def channel_infos(self, name):
        parsed = parse_telegram_url(name)

        if isinstance(parsed, ParsedTelegramGroup):
            raise TelegramInvalidTargetError

        if isinstance(parsed, ParsedTelegramChannel) or isinstance(
            parsed, ParsedTelegramMessage
        ):
            name = parsed.name

        url = convert_telegram_url_to_public(forge_telegram_channel_url(name))

        html = self.request_page(url)

        return scrape_channel_infos(html)

    def channel_messages(self, name):
        parsed = parse_telegram_url(name)

        if isinstance(parsed, ParsedTelegramGroup):
            raise TelegramInvalidTargetError

        if isinstance(parsed, ParsedTelegramChannel) or isinstance(
            parsed, ParsedTelegramMessage
        ):
            name = parsed.name

        url = convert_telegram_url_to_public(forge_telegram_channel_url(name))

        def generator():
            current_url = url

            while True:
                html = self.request_page(current_url)

                next_after, messages = scrape_channel_messages(html)

                for message in messages:
                    yield message

                if next_after is None:
                    break

                current_url = forge_telegram_channel_url_next(url, next_after)

        return generator()



# =============================================================================
# Minet Twitter Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class TwitterError(MinetError):
    pass


class TwitterGuestTokenError(TwitterError):
    pass


class TwitterPublicAPIQueryTooLongError(TwitterError):
    pass


class TwitterPublicAPIRateLimitError(TwitterError):
    pass


class TwitterPublicAPIInvalidResponseError(TwitterError):
    pass


class TwitterPublicAPIBadRequest(TwitterError):
    pass


class TwitterPublicAPIOverCapacityError(TwitterError):
    pass


class TwitterPublicAPIParsingError(TwitterError):
    pass


class TwitterPublicAPIHiccupError(TwitterError):
    pass



# =============================================================================
# Minet CrowdTangle Search
# =============================================================================
#
# Function related to post search.
#
from urllib.parse import quote

from minet.crowdtangle.utils import make_paginated_iterator
from minet.crowdtangle.formatters import format_post

URL_TEMPLATE = "https://api.crowdtangle.com/posts/search?count=100&sortBy=%(sort_by)s&token=%(token)s&searchTerm=%(terms)s"


def url_forge(**kwargs):
    base_url = URL_TEMPLATE % {
        "sort_by": kwargs["sort_by"],
        "token": kwargs["token"],
        "terms": quote(kwargs["terms"]),
    }

    if kwargs.get("start_date") is not None:
        base_url += "&startDate=%s" % kwargs["start_date"]

    if kwargs.get("end_date") is not None:
        base_url += "&endDate=%s" % kwargs["end_date"]

    if kwargs.get("platforms") is not None:
        base_url += "&platforms=%s" % ",".join(kwargs["platforms"])

    if kwargs.get("types") is not None:
        base_url += "&types=%s" % ",".join(kwargs["types"])

    if kwargs.get("offset") is not None:
        base_url += "&offset=%s" % kwargs["offset"]

    if kwargs.get("not_in_title") is not None:
        base_url += "&notInTitle"

    if kwargs.get("and") is not None:
        base_url += "&and=%s" % quote(kwargs["and"])

    if kwargs.get("language") is not None:
        base_url += "&language=%s" % kwargs["language"]

    if kwargs.get("search_field") is not None:
        base_url += "&searchField=%s" % kwargs["search_field"]

    if kwargs.get("in_list_ids") is not None:
        base_url += "&inListIds=%s" % (",".join(kwargs["in_list_ids"]))

    return base_url


crowdtangle_search = make_paginated_iterator(
    url_forge, item_key="posts", formatter=format_post
)



# =============================================================================
# Minet Crawler
# =============================================================================
#
# Crawler class invoking multiple spiders to scrape data from the web.
#
from typing import (
    cast,
    Optional,
    TypeVar,
    Callable,
    Dict,
    Tuple,
    Any,
    Generic,
    Mapping,
    Iterable,
    Iterator,
    Union,
)

from minet.types import AnyFileTarget
from minet.fs import load_definition
from minet.crawl.types import (
    CrawlJob,
    UrlOrCrawlJob,
    CrawlJobDataType,
    CrawlJobOutputDataType,
    CrawlResult,
)
from minet.crawl.spiders import (
    Spider,
    FunctionSpider,
    FunctionSpiderCallable,
    DefinitionSpider,
)
from minet.crawl.queue import CrawlerQueue, DumpType
from minet.crawl.state import CrawlerState
from minet.web import request, EXPECTED_WEB_ERRORS, AnyTimeout
from minet.fetch import HTTPThreadPoolExecutor, CANCELLED
from minet.exceptions import UnknownSpiderError, CancelledRequestError
from minet.constants import (
    DEFAULT_DOMAIN_PARALLELISM,
    DEFAULT_IMAP_BUFFER_SIZE,
    DEFAULT_THROTTLE,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_URLLIB3_TIMEOUT,
)

DEFAULT_SPIDER_KEY = "$$DEFAULT_MINET_SPIDER$$"


def ensure_job(
    url_or_job: UrlOrCrawlJob[CrawlJobDataType],
) -> CrawlJob[CrawlJobDataType]:
    if isinstance(url_or_job, CrawlJob):
        return url_or_job

    return CrawlJob(url=url_or_job)


RequestArgsType = Callable[[CrawlJob[CrawlJobDataType]], Dict]


class CrawlWorker(Generic[CrawlJobDataType, CrawlJobOutputDataType]):
    def __init__(
        self,
        crawler: "Crawler",
        *,
        request_args: Optional[RequestArgsType[CrawlJobDataType]] = None,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        callback: Optional[
            Callable[[CrawlResult[CrawlJobDataType, CrawlJobOutputDataType]], None]
        ] = None,
    ):
        self.crawler = crawler

        self.request_args = request_args
        self.max_redirects = max_redirects
        self.callback = callback

        self.cancel_event = self.crawler.executor.cancel_event
        self.local_context = self.crawler.executor.local_context

        self.default_request_kwargs = {
            "pool_manager": crawler.executor.pool_manager,
            "max_redirects": max_redirects,
            "cancel_event": crawler.executor.cancel_event,
        }

    def __call__(
        self, job: CrawlJob[CrawlJobDataType]
    ) -> Union[object, CrawlResult[CrawlJobDataType, CrawlJobOutputDataType]]:

        # Registering work
        with self.crawler.state.task():
            cancel_event = self.cancel_event

            result = CrawlResult(job)
            spider = self.crawler.get_spider(job.spider)

            if spider is None:
                result.error = UnknownSpiderError(spider=job.spider)
                return result

            # NOTE: crawl job must have a url and a depth at that point
            assert job.url is not None
            assert job.depth is not None

            kwargs = {}

            if cancel_event.is_set():
                return CANCELLED

            if self.request_args is not None:
                # NOTE: request_args must be threadsafe
                kwargs = self.request_args(job)

            if cancel_event.is_set():
                return CANCELLED

            try:
                retryer = getattr(self.local_context, "retryer", None)
                kwargs.update(self.default_request_kwargs)

                if retryer is not None:
                    response = retryer(request, job.url, **kwargs)
                else:
                    response = request(job.url, **kwargs)

            except CancelledRequestError:
                return CANCELLED

            except EXPECTED_WEB_ERRORS as error:
                result.error = error
                return result

            finally:
                job.attempts += 1

            result.response = response

            if cancel_event.is_set():
                return CANCELLED

            try:
                output, next_jobs = spider(job, response)
                result.output = output

                if cancel_event.is_set():
                    return CANCELLED

                if next_jobs is not None:
                    result.degree = self.crawler.enqueue(
                        next_jobs, spider=job.spider, depth=job.depth + 1
                    )

            except Exception as error:
                result.error = error

            return result


CrawlJobDataTypes = TypeVar("CrawlJobDataTypes", bound=Mapping)
CrawlJobOutputDataTypes = TypeVar("CrawlJobOutputDataTypes")
Spiders = Union[
    Spider[CrawlJobDataTypes, CrawlJobOutputDataTypes],
    Dict[str, Spider[CrawlJobDataTypes, CrawlJobOutputDataTypes]],
]

# TODO: try creating a kwarg type for those
# NOTE: crawling could work depth-first if we wanted
class Crawler(Generic[CrawlJobDataTypes, CrawlJobOutputDataTypes]):
    executor: HTTPThreadPoolExecutor
    queue: CrawlerQueue[CrawlJob[CrawlJobDataTypes]]
    persistent: bool
    state: CrawlerState
    started: bool
    stopped: bool
    resuming: bool
    finished: bool
    singular: bool

    __spiders: Dict[str, Spider[CrawlJobDataTypes, CrawlJobOutputDataTypes]]

    def __init__(
        self,
        spiders: Spiders[CrawlJobDataTypes, CrawlJobOutputDataTypes],
        queue_path: Optional[str] = None,
        resume: bool = False,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
        throttle: float = DEFAULT_THROTTLE,
        max_workers: Optional[int] = None,
        wait: bool = True,
        daemonic: bool = False,
        insecure: bool = False,
        timeout: AnyTimeout = DEFAULT_URLLIB3_TIMEOUT,
        spoof_tls_ciphers: bool = False,
        proxy: Optional[str] = None,
        retry: bool = False,
        retryer_kwargs: Optional[Dict[str, Any]] = None,
    ):

        # Own executor and imap params
        self.executor = HTTPThreadPoolExecutor(
            max_workers=max_workers,
            insecure=insecure,
            timeout=timeout,
            wait=wait,
            daemonic=daemonic,
            spoof_tls_ciphers=spoof_tls_ciphers,
            proxy=proxy,
            retry=retry,
            retryer_kwargs=retryer_kwargs,
        )

        self.imap_kwargs = {
            "buffer_size": buffer_size,
            "parallelism": domain_parallelism,
            "throttle": throttle,
        }

        # Params
        self.queue_path = queue_path
        self.persistent = queue_path is not None

        # Lifecycle
        self.started = False
        self.stopped = False
        self.resuming = False
        self.finished = False

        # Queue
        self.queue = CrawlerQueue(queue_path, resume=resume, cleanup_interval=2)
        self.persistent = self.queue.persistent
        self.resuming = self.queue.resuming

        if self.resuming and self.queue.qsize() == 0:
            self.finished = True

        # Initializing state
        self.state = CrawlerState(jobs_queued=self.queue.qsize())

        # Spiders
        if isinstance(spiders, Spider):
            self.__spiders = {DEFAULT_SPIDER_KEY: spiders}
            self.singular = True
        elif isinstance(spiders, Mapping):
            self.__spiders = {}
            self.singular = False

            for name, spider in spiders.items():
                self.__spiders[name] = spider
        else:
            raise TypeError("expecting a single spider or a mapping of spiders")

    def __repr__(self):
        class_name = self.__class__.__name__

        return "<{class_name} {number}>".format(
            class_name=class_name, number="singular" if self.singular else "plural"
        )

    @property
    def plural(self) -> bool:
        return not self.singular

    def get_spider(self, name: Optional[str] = None) -> Spider:
        if name is None and self.plural:
            raise TypeError("plural crawler cannot return default spider")

        if name is not None and self.singular:
            raise TypeError("singular crawler cannot return a spider by name")

        if name is None:
            name = DEFAULT_SPIDER_KEY

        return self.__spiders[name]

    def spiders(self) -> Iterator[Tuple[str, Spider]]:
        if self.singular:
            raise TypeError(
                "singular crawler cannot iterate over its spiders (it only has the default one, use #.get_spider to get it)"
            )

        for name, spiders in self.__spiders.items():
            yield name, spiders

    def start(self) -> None:
        # TODO: maybe start the executor ourselves if required when quenouille moves

        if self.stopped:
            # TODO: we could but we need to mind the condition below
            raise RuntimeError("Cannot restart a crawler")

        if self.finished:
            raise RuntimeError("Cannot start an already finished crawler")

        if self.started:
            raise RuntimeError("Crawler has already started")

        # Collecting start jobs - we only add those if we are not resuming
        if not self.resuming:

            # NOTE: start jobs are all buffered into memory
            # We could use a blocking queue with max size but this could prove
            # difficult to resume crawls based upon lazy iterators
            for name, spider in self.__spiders.items():
                spider_start_jobs = spider.start_jobs()

                if self.singular:
                    name = None

                if spider_start_jobs is not None:
                    self.enqueue(spider_start_jobs, spider=name)

        self.started = True

    def stop(self):
        self.stopped = True
        self.executor.shutdown(wait=self.executor.wait)
        del self.queue

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc):
        self.stop()

    def __iter__(
        self,
    ) -> Iterator[CrawlResult[CrawlJobDataTypes, CrawlJobOutputDataTypes]]:
        worker = CrawlWorker(self)

        def key_by_domain_name(job: CrawlJob) -> Optional[str]:
            return job.domain

        imap_unordered = self.executor.imap_unordered(
            self.queue, worker, key=key_by_domain_name, **self.imap_kwargs
        )

        def safe_wrapper():
            for result in imap_unordered:
                if result is CANCELLED:
                    continue

                yield result

                self.queue.ack(result.job)

            # If iterator ended properly we cleanup the queue
            self.queue.cleanup()

        return safe_wrapper()

    def enqueue(
        self,
        job_or_jobs: Union[
            UrlOrCrawlJob[CrawlJobDataTypes], Iterable[UrlOrCrawlJob[CrawlJobDataTypes]]
        ],
        spider: Optional[str] = None,
        depth: int = 0,
    ) -> int:
        if isinstance(job_or_jobs, (str, CrawlJob)):
            jobs = [job_or_jobs]
        else:
            jobs = job_or_jobs

        def proper_jobs():
            for job in jobs:
                job = ensure_job(job)

                if spider is not None and job.spider is None:
                    job.spider = spider

                if job.depth is None:
                    job.depth = depth

                yield job

        count = self.queue.put_many(proper_jobs())

        self.state.inc_queued(count)

        return count

    # NOTE: this is clearly not threadsafe lol. This is for debug only.
    def dump_queue(self) -> DumpType[CrawlJob[CrawlJobDataTypes]]:
        if self.started:
            raise TypeError("cannot dump queue while crawler is running")

        return self.queue.dump()

    @classmethod
    def from_callable(
        cls,
        fn: FunctionSpiderCallable,
        start_jobs: Optional[Iterable[UrlOrCrawlJob]] = None,
        **kwargs,
    ):
        return cls(FunctionSpider(fn, start_jobs=start_jobs), **kwargs)

    @classmethod
    def from_definition(
        cls, definition: Union[Dict[str, Any], AnyFileTarget], **kwargs
    ):
        if not isinstance(definition, dict):
            definition = load_definition(definition)

        definition = cast(Dict[str, Any], definition)

        # Do we have a single spider or multiple spiders?
        if "spiders" in definition:
            spiders = {
                name: DefinitionSpider(s) for name, s in definition["spiders"].items()
            }
        else:
            spiders = DefinitionSpider(definition)

        return cls(spiders, **kwargs)



# =============================================================================
# Minet Twitter API Client
# =============================================================================
#
# Twitter public API client.
#
from json import JSONDecodeError
from twitwi import TwitterWrapper
from twitwi.exceptions import TwitterWrapperMaxAttemptsExceeded
from twitter import TwitterError

from minet.web import create_request_retryer, retrying_method


def retryer_predicate(exc):
    if isinstance(exc, (JSONDecodeError, TwitterWrapperMaxAttemptsExceeded)):
        return True

    if not isinstance(exc, TwitterError):
        return False

    msg = str(exc).lower()

    # Retrying if we actually caught a json decoding error
    # TODO: simplify this if the `twitter` lib changes its error inheritance scheme
    if "incomplete json data collected" in msg:
        return True

    return False


class TwitterAPIClient(object):
    def __init__(
        self,
        access_token,
        access_token_secret,
        api_key,
        api_secret_key,
        api_version="1.1",
    ):
        self.wrapper = TwitterWrapper(
            access_token,
            access_token_secret,
            api_key,
            api_secret_key,
            api_version=api_version,
        )
        self.retryer = create_request_retryer(predicate=retryer_predicate)

    @retrying_method()
    def call(self, *args, **kwargs):
        return self.wrapper.call(*args, **kwargs)



# =============================================================================
# Minet Default Values Collection
# =============================================================================
#
# Listing sane default values used throughout the whole library.
#
from urllib3 import Timeout


# Fetch-related
DEFAULT_DOMAIN_PARALLELISM = 1
DEFAULT_IMAP_BUFFER_SIZE = 1024
DEFAULT_THROTTLE = 0.2
DEFAULT_THREADS = 25
DEFAULT_CONNECT_TIMEOUT = 5
DEFAULT_FETCH_MAX_REDIRECTS = 5
DEFAULT_RESOLVE_MAX_REDIRECTS = 20
DEFAULT_READ_TIMEOUT = 25
DEFAULT_URLLIB3_TIMEOUT = Timeout(
    connect=DEFAULT_CONNECT_TIMEOUT, read=DEFAULT_READ_TIMEOUT
)
DEFAULT_SPOOFED_UA = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:87.0) Gecko/20100101 Firefox/87.0"
)
DEFAULT_SPOOFED_TLS_CIPHERS = "TLS_AES_128_GCM_SHA256:TLS_AES_256_GCM_SHA384:TLS_CHACHA20_POLY1305_SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA:AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA:AES256-SHA"

COOKIE_BROWSERS = {"chrome", "chromium", "firefox", "opera", "edge"}



# =============================================================================
# Minet CrowdTangle Formatters
# =============================================================================
#
# Various data formatters for CrowdTangle API data.
#
from casanova import namedrecord

from minet.crowdtangle.constants import (
    CROWDTANGLE_POST_TYPES,
    CROWDTANGLE_REACTION_TYPES,
    CROWDTANGLE_POST_CSV_HEADERS,
    CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK,
    CROWDTANGLE_SUMMARY_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN,
    CROWDTANGLE_LIST_CSV_HEADERS,
    CROWDTANGLE_STATISTICS,
    CROWDTANGLE_FULL_STATISTICS,
)

CrowdTanglePost = namedrecord(
    "CrowdTanglePost",
    CROWDTANGLE_POST_CSV_HEADERS,
    boolean=["account_verified"],
    plural=["links", "expanded_links"],
    json=["media"],
)

CrowdTanglePostWithLink = namedrecord(
    "CrowdTanglePostWithLink",
    CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK,
    boolean=["account_verified"],
    plural=["links", "expanded_links"],
    json=["media"],
)

CrowdTangleSummary = namedrecord("CrowdTangleSummary", CROWDTANGLE_SUMMARY_CSV_HEADERS)

CrowdTangleLeaderboard = namedrecord(
    "CrowdTangleLeaderboard", CROWDTANGLE_LEADERBOARD_CSV_HEADERS, boolean=["verified"]
)

CrowdTangleLeaderboardWithBreakdown = namedrecord(
    "CrowdTangleLeaderboardWithBreakdown",
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN,
    boolean=["verified"],
)

CrowdTangleList = namedrecord("CrowdTangleList", CROWDTANGLE_LIST_CSV_HEADERS)


def map_key(key, target):
    return [item[key] for item in target]


def format_post(post, link=None):
    row = [
        post["id"],
        post["platformId"],
        post["platform"],
        post["type"],
        post.get("title"),
        post.get("caption"),
        post.get("message"),
        post.get("description"),
        post["date"].split(" ", 1)[0],
        post["date"],
        post["updated"],
        post.get("link"),
        post.get("postUrl"),
        post["score"],
        post.get("videoLengthMS"),
        post.get("liveVideoStatus"),
    ]

    if link:
        row = [link] + row

    stats = post["statistics"]
    actual_stats = stats["actual"]
    expected_stats = stats["expected"]

    for name in CROWDTANGLE_STATISTICS:
        key = "%sCount" % name

        row.append(actual_stats.get(key, ""))
        row.append(expected_stats.get(key, ""))

    account = post["account"]

    row.extend(
        [
            # Account
            account["id"],
            account.get("platformId"),
            account.get("platform"),
            account["name"],
            account.get("handle"),
            account.get("profileImage"),
            account["subscriberCount"],
            account["url"],
            account["verified"],
            account.get("accountType"),
            account.get("pageAdminTopCountry"),
            # Remaining
            map_key("original", post.get("expandedLinks", [])),
            map_key("expanded", post.get("expandedLinks", [])),
            post.get("media"),
        ]
    )

    if link is not None:
        return CrowdTanglePostWithLink(*row)

    return CrowdTanglePost(*row)


def format_summary(stats):
    row = (stats["%sCount" % t] for t in CROWDTANGLE_REACTION_TYPES)
    return CrowdTangleSummary(*row)


def format_leaderboard(item, with_breakdown=False):
    account = item["account"]
    subscriber_data = item["subscriberData"]

    row = [
        account["id"],
        account["name"],
        account.get("handle"),
        account.get("profileImage"),
        account["subscriberCount"],
        account["url"],
        account["verified"],
        subscriber_data["initialCount"],
        subscriber_data["finalCount"],
        subscriber_data.get("notes"),
    ]

    summary = item["summary"]

    for key, _ in CROWDTANGLE_FULL_STATISTICS:
        row.append(summary.get(key))

    if with_breakdown:
        breakdown = item["breakdown"]

        for post_type in CROWDTANGLE_POST_TYPES:

            data = breakdown.get(post_type)

            for key, _ in CROWDTANGLE_FULL_STATISTICS:
                row.append(data.get(key) if data else None)

    if with_breakdown:
        return CrowdTangleLeaderboardWithBreakdown(*row)

    return CrowdTangleLeaderboard(*row)


def format_list(item):
    return CrowdTangleList(item["id"], item["title"], item["type"])



# =============================================================================
# Minet Loading Bars
# =============================================================================
#
# Various loading bar utilities used by minet CLI.
#
from typing import Optional, Iterable
from typing_extensions import TypedDict, NotRequired

from contextlib import contextmanager
from collections import OrderedDict, namedtuple
from rich.live import Live
from rich.text import Text
from rich.table import Table, Column
from rich.progress import (
    Progress,
    ProgressColumn,
    TextColumn,
    BarColumn,
    SpinnerColumn,
    Task,
)
from about_time import HumanDuration, HumanThroughput
from ebbe import format_int

from minet.utils import message_flatmap
from minet.cli.console import console


class CautiousBarColumn(BarColumn):
    def render(self, task: Task):
        if task.total is None:
            self.bar_width = 15
        else:
            self.bar_width = min(40, 10 + task.total)

        if task.total is not None and task.completed > task.total:
            self.finished_style = "warning"

        return super().render(task)


class TimeElapsedColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed

        if not elapsed:
            return Text("")

        return Text(
            "in " + HumanDuration(elapsed).as_human(2), style="progress.elapsed"
        )


class CompletionColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        total = format_int(task.total) if task.total is not None else "?"
        completed = format_int(task.completed).rjust(len(total))

        unit = task.fields.get("unit")
        unit_text = f" {unit}" if unit is not None else ""

        return Text(f"{completed}/{total}{unit_text}")


class PercentageColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        if task.total is None or task.total == 1:
            return Text("-")

        txt = "[progress.percentage][{task.percentage:>3.0f}%]".format(task=task)
        return Text.from_markup(txt)


class ThroughputColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed

        if not elapsed or not task.completed:
            return Text("(?/s)")

        throughput = HumanThroughput(task.completed / elapsed, "").as_human(2)

        eta = ""

        if task.total is not None and task.time_remaining:
            remaining = HumanDuration(task.time_remaining).as_human(2)
            eta = f" eta: {remaining}"

        message = f"({throughput}{eta})"

        return Text(message)


class StatsColumn(ProgressColumn):
    def __init__(self, table_column=None, sort_key=None):
        super().__init__(table_column=table_column)
        self.sort_key = sort_key

    def render(self, task: Task) -> Text:
        stats = task.fields.get("stats")

        if stats is None:
            return Text("")

        total = 0

        item_parts = []

        values = stats.values()

        if self.sort_key is not None:
            values = sorted(values, key=self.sort_key)

        for item in values:
            txt = Text()

            count = item["count"]

            if count == 0:
                continue

            total += count

            txt.append(str(item["name"]), style=item["style"])
            txt.append(" ")
            txt.append(format_int(count))

            item_parts.append(txt)

        parts = []

        if task.description and total:
            parts.append(Text("- "))

        return Text.assemble(*parts, *Text(" ").join(item_parts))


class NestedTotalColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        sub_total_sum = task.fields.get("sub_total_sum")

        if sub_total_sum is None:
            return None

        unit = task.fields.get("unit")
        unit_text = f" {unit}" if unit is not None else ""

        return Text(f"total: {format_int(sub_total_sum)}{unit_text}")


class StatsItem(TypedDict):
    name: str
    style: NotRequired[str]
    initial: NotRequired[int]


CatchAttempt = namedtuple("CatchAttempt", ["item", "index", "exception"])


def attempt_to_catch(catch, exc, item, index=None) -> bool:
    if catch is None:
        return True

    msg = None

    # Callable version
    if callable(catch):
        return catch(CatchAttempt(item, index, exc))

    # Mapping version
    msg_format = catch.get(type(exc))

    if msg_format is None:
        return True

    if msg_format is not None:
        msg = msg_format.format(item=item, index=index)

    if index is not None:
        console.logh("Line %i" % (index + 1), msg)
    else:
        console.print(msg)

    return False


class LoadingBar(object):
    def __init__(
        self,
        title: Optional[str] = None,
        unit: Optional[str] = None,
        total: Optional[int] = None,
        completed: Optional[int] = None,
        show_label: bool = False,
        label_format: str = "{task.description}",
        stats: Optional[Iterable[StatsItem]] = None,
        stats_sort_key=None,
        nested: bool = False,
        sub_title: Optional[str] = None,
        sub_title_style: str = "info",
        sub_unit: Optional[str] = None,
        transient: bool = False,
    ):
        self.sub_total_sum = 0
        self.nested = nested
        self.show_label = show_label
        self.transient = transient
        self.known_total = total is not None
        self.already_stopped = False

        self.bar_column = None
        self.label_progress = None
        self.label_progress_task_id = None
        self.sub_progress = None
        self.sub_task_id = None
        self.stats_progress = None
        self.stats_task = None
        self.stats = OrderedDict()
        self.stats_are_shown = False

        self.table = Table.grid(expand=True)

        # Label line
        if show_label:
            label_progress_columns = []

            if show_label:
                label_progress_columns.append(TextColumn(label_format))

            self.label_progress = Progress(*label_progress_columns)
            self.label_progress_task_id = self.label_progress.add_task(
                description="", total=None
            )
            self.table.add_row(self.label_progress)

        # Main progress line
        self.bar_column = CautiousBarColumn(pulse_style="white")

        columns = [
            TextColumn("[progress.description]{task.description}"),
            self.bar_column,
        ]

        columns.append(CompletionColumn(Column(overflow="ellipsis", no_wrap=True)))

        if not nested:
            columns.append(SpinnerColumn("dots", style=None, finished_text="·"))

        columns.append(PercentageColumn())
        columns.append(TimeElapsedColumn())
        columns.append(ThroughputColumn(Column(overflow="ellipsis", no_wrap=True)))

        self.progress = Progress(*columns)
        self.table.add_row(self.progress)

        self.task_id = self.progress.add_task(
            description=title or "",
            total=total,
            unit=unit,
            completed=completed or 0,
        )

        # Nested progress line
        if nested:
            sub_columns = [
                SpinnerColumn("dots", style="", finished_text="·"),
                TextColumn(
                    "{task.description}",
                    table_column=Column(max_width=50),
                    style=sub_title_style,
                ),
                CompletionColumn(),
                TimeElapsedColumn(),
                ThroughputColumn(),
                NestedTotalColumn(),
            ]

            self.sub_progress = Progress(*sub_columns)
            self.sub_task_id = self.sub_progress.add_task(
                description=sub_title or "",
                total=None,
                sub_total_sum=self.sub_total_sum,
                unit=sub_unit,
            )

            self.table.add_row(self.sub_progress)

        # Stats line
        if stats is not None:
            for item in stats:
                count = item.get("count", 0)

                self.stats[item["name"]] = {
                    "name": item["name"],
                    "style": item.get("style", ""),
                    "count": count,
                }

                if count:
                    self.stats_are_shown = True

        self.stats_progress = Progress(StatsColumn(sort_key=stats_sort_key))
        self.stats_task_id = self.stats_progress.add_task("", stats=self.stats)

        if self.stats_are_shown:
            self.table.add_row(self.stats_progress)

        # Internal live instance
        self.live = Live(
            self.table, refresh_per_second=10, console=console, transient=self.transient
        )

    def cursor_up(self) -> None:
        # NOTE: cursor 1up
        console.file.write("\x1b[1A")

    def start(self) -> None:
        self.live.start()

    def stop(self, erase=False) -> None:
        if erase:
            self.live.transient = True

        self.live.stop()
        self.already_stopped = True

    def erase(self) -> None:
        self.stop(erase=True)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):

        if exc_type is not None:
            style = "error"

            if exc_type is KeyboardInterrupt:
                if not self.already_stopped:
                    self.cursor_up()

                style = "warning"

                # NOTE: broken pipe often arrive after that point

            if self.bar_column is not None:
                self.bar_column.complete_style = style
                self.bar_column.finished_style = style
                self.bar_column.pulse_style = style

                if not self.known_total:
                    self.bar_column.style = style
        else:
            self.bar_column.pulse_style = "success"
            self.bar_column.style = "success"

        self.stop()

    @contextmanager
    def step(self, item=None, count=1, index=None, catch=None, sub_total=None):
        interrupted = False

        # TODO: we could add some key kwarg if needed

        try:
            if self.nested:
                self.nested_reset()

                if item is not None:
                    self.update(sub_title=item)

                self.set_sub_total(sub_total)
            else:
                if self.show_label and item is not None:
                    self.set_label(item)
            yield

        except BaseException as exc:
            interrupted = attempt_to_catch(catch, exc, item, index)

            if interrupted:
                raise
        finally:
            if not interrupted:
                self.advance(count)

    @contextmanager
    def nested_step(self, count=1):
        assert self.nested
        interrupted = False

        try:
            yield
        except BaseException:
            interrupted = True
            raise
        finally:
            if not interrupted:
                self.nested_advance(count)

    def advance(self, count=1):
        self.progress.update(self.task_id, advance=count)

    def nested_advance(self, count=1):
        self.sub_total_sum += 1
        self.sub_progress.update(
            self.sub_task_id, advance=count, sub_total_sum=self.sub_total_sum
        )

    def nested_reset(self):
        self.sub_progress.reset(self.sub_task_id)

    def __refresh_stats(self):
        self.stats_progress.update(self.stats_task_id, stats=self.stats)

        if not self.stats_are_shown:
            self.stats_are_shown = True
            self.table.add_row(self.stats_progress)

    def set_title(self, title: str):
        self.progress.update(self.task_id, description=title)

    def set_total(self, total: Optional[int] = None):
        self.progress.update(self.task_id, total=total)
        self.known_total = total is not None

    def set_sub_total(self, total: Optional[int] = None):
        self.sub_progress.update(self.sub_task_id, total=total)

    def set_label(self, label: str):
        assert self.label_progress is not None
        self.label_progress.update(self.label_progress_task_id, description=label)

    # TODO: factorize
    def set_stat(self, name: str, count: int, style: str = None):
        assert self.stats is not None

        if name not in self.stats:
            self.stats[name] = {"name": name, "count": count, "style": style or ""}
        else:
            item = self.stats[name]
            item["count"] = count

            if style is not None:
                item["style"] = style

        self.__refresh_stats()

    def inc_stat(self, name: str, count: int = 1, style: Optional[str] = None):
        assert self.stats is not None

        if name not in self.stats:
            self.stats[name] = {"name": name, "count": count, "style": style or ""}
        else:
            item = self.stats[name]
            item["count"] += count

            if style is not None:
                item["style"] = style

        self.__refresh_stats()

    def update(self, count=None, sub_title=None, sub_count=None, label=None, **fields):
        if count is not None:
            self.advance(count)

        if sub_count is not None:
            self.nested_advance(sub_count)

        if sub_title is not None:
            self.sub_progress.update(self.sub_task_id, description=sub_title)

        if label is not None:
            self.set_label(label)

        if fields:
            assert self.stats is not None

            for field, count in fields.items():
                self.stats[field]["count"] = count

            self.__refresh_stats()

    def print(self, *msg, end="\n"):
        console.print(message_flatmap(*msg), end=end)

    def error(self, *args, **kwargs):
        console.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        console.warning(*args, **kwargs)



# =============================================================================
# Minet Argparse Helpers
# =============================================================================
#
# Miscellaneous helpers related to CLI argument parsing.
#
from typing import Optional
from typing_extensions import TypedDict, NotRequired

import os
import re
import sys
import shutil
from os.path import isdir
from io import TextIOBase
from argparse import (
    Action,
    ArgumentError,
    ArgumentTypeError,
    ArgumentParser,
)
from rich_argparse import RawDescriptionRichHelpFormatter
from gettext import gettext
from textwrap import dedent
from casanova import Resumer, CsvCellIO
from ebbe import getpath, omit
from datetime import datetime
from pytz import timezone
from pytz.exceptions import UnknownTimeZoneError

from minet.cli.console import MINET_COLORS
from minet.cli.exceptions import NotResumableError, InvalidArgumentsError
from minet.cli.utils import acquire_cross_platform_stdout

TEMPLATE_RE = re.compile(r"<%\s+([A-Za-z/\-]+)\s+%>")

ARGUMENT_PREFIXES_TO_NORMALIZE = ["--dont-", "--no-", "--", "-"]
FLAG_SORTING_PRIORITIES = {
    name: i
    for i, name in enumerate(
        [
            "select",
            "explode",
            "total",
            "input",
            "output",
            "resume",
            "rcfile",
            "silent",
            "help",
        ],
        1,
    )
}

# NOTE: custom style for rich_argparse
RawDescriptionRichHelpFormatter.styles["argparse.groups"] = MINET_COLORS["warning"]
RawDescriptionRichHelpFormatter.styles["argparse.dim"] = "dim"
RawDescriptionRichHelpFormatter.styles["argparse.emphasis"] = (
    "italic " + MINET_COLORS["warning"]
)
RawDescriptionRichHelpFormatter.styles["argparse.title"] = (
    "bold " + MINET_COLORS["error"]
)

# NOTE: custom highlighting for rich_argparse
RawDescriptionRichHelpFormatter.highlights = [
    r"(?P<args>-[a-zA-Z])[\s/.]",  # -f flags
    r"(?P<args>--[a-z]+(-[a-z]+)*)[\s/.]",  # --flag flags
    r"(?P<title>^#\s+.+)",  # command title
    r"(?P<title>BEWARE):",  # beware
    r"\s+\$\s+(?P<args>.+)",  # examples
    r"(?P<dim>\n>\s+.+)",  # caret sections
    r'"(?P<metavar>[^"]+)"',  # double-quote literals
    r"`(?P<metavar>[^`]+)`",  # backtick literals
    r"(?P<metavar><[a-z\s=\"]+>)",  # tags
    r"(?P<emphasis>how to use the command with.+)",  # emphasis
    r"(?P<args>https?://\S+)",  # urls
    r"(?P<groups>\n(?:--[a-z]|[A-Z]).+:\n)",  # various titles
]


def normalize_argument_name(name: str) -> str:
    for prefix in ARGUMENT_PREFIXES_TO_NORMALIZE:
        if name.startswith(prefix):
            return name[len(prefix) :]

    return name


def arguments_sort_key(option_strings):
    if not option_strings:
        return (0, 0)

    longest_name = max(option_strings, key=len)
    longest_name = normalize_argument_name(longest_name)

    # Special flags that are dragged down
    priority = FLAG_SORTING_PRIORITIES.get(longest_name, 0)

    return (priority, longest_name)


class SortingRawTextHelpFormatter(RawDescriptionRichHelpFormatter):
    def add_arguments(self, actions) -> None:
        actions = sorted(actions, key=lambda a: arguments_sort_key(a.option_strings))
        return super().add_arguments(actions)


def custom_formatter(prog):
    terminal_size = shutil.get_terminal_size()

    return SortingRawTextHelpFormatter(
        prog, width=terminal_size.columns, max_help_position=32
    )


def get_subparser(o, keys):
    parser = None

    for key in keys:
        item = o.get(key)

        if item is None:
            return None

        parser = item["parser"]

        if "subparsers" in item:
            o = item["subparsers"]
        else:
            break

    return parser


ARGUMENT_KEYS_TO_OMIT = ["name", "flag", "flags"]


def add_arguments(subparser, arguments):
    for argument in arguments:

        argument_kwargs = omit(argument, ARGUMENT_KEYS_TO_OMIT)

        if "choices" in argument_kwargs:
            argument_kwargs["choices"] = sorted(argument_kwargs["choices"])

        if "name" in argument:
            subparser.add_argument(argument["name"], **argument_kwargs)
        elif "flag" in argument:
            subparser.add_argument(argument["flag"], **argument_kwargs)
        else:
            subparser.add_argument(*argument["flags"], **argument_kwargs)

        if "action" in argument and argument["action"] is ConfigAction:
            try:
                subparser.add_argument(
                    "--rcfile",
                    help="Custom path to a minet configuration file. More info about this here: https://github.com/medialab/minet/blob/master/docs/cli.md#minetrc",
                )
            except ArgumentError:
                pass

        try:
            subparser.add_argument(
                "--silent",
                help="Whether to suppress all the log and progress bars. Can be useful when piping.",
                action="store_true",
            )
        except ArgumentError:
            pass


def build_description(command):
    description = "# " + command["title"]

    text = dedent(command.get("description", ""))
    description += "\n\n" + text

    return description


def build_subparsers(
    parser,
    index,
    commands,
    help="Action to execute",
    title="actions",
    dest="action",
    common_arguments=[],
):

    subparsers = parser.add_subparsers(help=help, title=title, dest=dest)

    for name, command in commands.items():
        subparser = subparsers.add_parser(
            name,
            description=build_description(command),
            epilog=dedent(command.get("epilog", "")),
            formatter_class=custom_formatter,
            aliases=command.get("aliases", []),
        )

        to_index = {"parser": subparser, "command": command, "subparsers": {}}

        add_arguments(subparser, common_arguments)

        if "arguments" in command:
            add_arguments(subparser, command["arguments"])

        if "subparsers" in command:
            subsubparsers = command["subparsers"]
            subcommon_arguments = subsubparsers.get("common_arguments", [])

            add_arguments(subparser, subcommon_arguments)

            build_subparsers(
                subparser,
                to_index["subparsers"],
                subsubparsers["commands"],
                help=subsubparsers["help"],
                title=subsubparsers["title"],
                dest=subsubparsers["dest"],
                common_arguments=common_arguments + subcommon_arguments,
            )

        if "aliases" in command:
            for alias in command["aliases"]:
                index[alias] = to_index

        index[name] = to_index

    return subparsers


def build_parser(name, version, commands):

    # Building the argument parser
    parser = ArgumentParser(prog="minet")

    parser.add_argument(
        "--version", action="version", version="%s %s" % (name, version)
    )

    subparser_index = {}

    subparsers = build_subparsers(
        parser, subparser_index, {c["name"]: c for c in commands}
    )

    # Help subparser
    help_subparser = subparsers.add_parser("help")
    help_subparser.add_argument("subcommand", help="Name of the subcommand", nargs="*")

    return parser, subparser_index


class TimestampType(object):
    def __call__(self, date):
        try:
            timestamp = int(datetime.strptime(date, "%Y-%m-%d").timestamp())
        except ValueError:
            raise ArgumentTypeError(
                "UTC date should have the following format : %Y-%m-%d"
            )
        return timestamp


class TimezoneType(object):
    def __call__(self, locale):
        try:
            tz = timezone(locale)
        except UnknownTimeZoneError:
            raise ArgumentTypeError("This timezone is not recognized.")
        return tz


class SplitterType(object):
    def __init__(self, splitchar=","):
        self.splitchar = splitchar

    def __call__(self, string):
        return string.split(self.splitchar)


class ExistingDirectoryType(object):
    def __call__(self, string):
        if not isdir(string):
            raise ArgumentTypeError('Could not find the "%s" directory!' % string)


class BooleanAction(Action):
    """
    Custom argparse action to handle --no-* flags.
    Taken from: https://thisdataguy.com/2017/07/03/no-options-with-argparse-and-python/
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

        if len(option_strings) == 2:
            self.true_options = [option_strings[0]]
            self.false_options = [option_strings[1]]
        elif len(option_strings) == 4:
            self.true_options = [option_strings[0], option_strings[1]]
            self.false_options = [option_strings[2], option_strings[3]]
        else:
            raise TypeError("expecting 2 or 4 flags")

    def __call__(self, parser, cli_args, values, option_string=None):
        setattr(
            cli_args,
            self.dest,
            False if option_string in self.false_options else True,
        )


class DummyCSVInput(object):
    ...


class SingleColumnDummyCSVInput(DummyCSVInput):
    def __init__(self, column, dest):
        self.column = column
        self.dest = dest

    def resolve(self, cli_args):
        value = getattr(cli_args, self.dest)

        if not value:
            raise InvalidArgumentsError("the command was given nothing!")

        f = CsvCellIO(value, column=self.column)
        setattr(cli_args, self.dest, self.column)
        setattr(cli_args, "has_dummy_csv", True)

        return f


class InputAction(Action):
    def __init__(
        self,
        option_strings,
        dest,
        dummy_csv_column=None,
        column_dest="column",
        default=None,
        **kwargs,
    ):
        if dummy_csv_column is not None:
            default = SingleColumnDummyCSVInput(dummy_csv_column, column_dest)

        super().__init__(option_strings, dest, default=default, **kwargs)

    def __call__(self, parser, cli_args, value, option_string=None):
        setattr(cli_args, "has_dummy_csv", False)

        if value == "-":
            f = sys.stdin
        else:
            try:
                f = open(value, "r", encoding="utf-8")
            except OSError as e:
                args = {"filename": value, "error": e}
                message = gettext("can't open '%(filename)s': %(error)s")
                raise ArgumentError(self, message % args)

        setattr(cli_args, self.dest, f)


class OutputOpener(object):
    def __init__(self, path, resumer_class=None, resumer_kwargs={}):
        self.path = path
        self.resumer_class = resumer_class
        self.resumer_kwargs = resumer_kwargs

    def open(self, cli_args, resume=False):
        if self.path == "-":
            if resume:
                raise NotResumableError

            return acquire_cross_platform_stdout()

        if resume and self.resumer_class is not None:
            resumer_kwargs = self.resumer_kwargs

            if callable(self.resumer_kwargs):
                resumer_kwargs = self.resumer_kwargs(cli_args)

            return self.resumer_class(self.path, **resumer_kwargs)

        mode = "a" if resume else "w"

        # As per #254: newline='' is necessary for CSV output on windows to avoid
        # outputting extra lines because of a '\r\r\n' end of line...
        return open(self.path, mode, encoding="utf-8", newline="")


class OutputAction(Action):
    def __init__(
        self,
        option_strings,
        dest,
        resumer=None,
        resumer_kwargs={},
        help="Path to the output file. Will consider `-` as stdout. If not given, results will also be printed to stdout.",
        default="-",
        **kwargs,
    ):
        self.resumer = resumer
        self.resumer_kwargs = resumer_kwargs
        super().__init__(
            option_strings,
            dest,
            help=help,
            default=OutputOpener(
                path=default, resumer_class=resumer, resumer_kwargs=resumer_kwargs
            ),
            **kwargs,
        )

    def __call__(self, parser, cli_args, value, option_string=None):
        opener = OutputOpener(
            path=value,
            resumer_class=self.resumer,
            resumer_kwargs=self.resumer_kwargs,
        )
        setattr(cli_args, self.dest, opener)


def rc_key_to_env_var(key):
    return "MINET_%s" % "_".join(token.upper() for token in key)


class WrappedConfigValue(object):
    def __init__(self, flag, key, default, _type, required=False):
        self.flag = flag
        self.key = key
        self.default = default
        self.type = _type
        self.required = required

    def resolve(self, config):
        value = None

        # Attempting to resolve env variable
        env_var = rc_key_to_env_var(self.key)
        env_value = os.environ.get(env_var, "").strip()

        if env_value:
            value = self.type(env_value)
        else:
            value = getpath(config, self.key, self.default)

        if value is None and self.required:
            raise InvalidArgumentsError(
                '%s is mandatory!\nIt can also be given using the %s env variable or using the "%s" .minetrc config key.\nMore info about this here: https://github.com/medialab/minet/blob/master/docs/cli.md#minetrc'
                % ("/".join(self.flag), env_var, ".".join(self.key))
            )

        return value


class ConfigAction(Action):
    def __init__(
        self,
        option_strings,
        dest,
        rc_key,
        default=None,
        plural=False,
        required=False,
        **kwargs,
    ):
        if "help" in kwargs:
            kwargs["help"] = kwargs["help"].rstrip(
                "."
            ) + '. Can also be configured in a .minetrc file as "%s" or read from the %s env variable.' % (
                ".".join(rc_key),
                rc_key_to_env_var(rc_key),
            )

        self.plural = plural

        if plural == True:
            self.list_values = []

        super().__init__(
            option_strings,
            dest,
            default=WrappedConfigValue(
                option_strings,
                rc_key,
                default,
                kwargs.get("type", str),
                required=required,
            ),
            **kwargs,
        )

    def __call__(self, parser, cli_args, values, option_string=None):
        if self.plural:
            self.list_values.append(values)
            setattr(cli_args, self.dest, self.list_values)
        else:
            setattr(cli_args, self.dest, values)


def resolve_arg_dependencies(cli_args, config):
    to_close = []

    # Unwrapping values
    # NOTE: I copy the dict from vars because we are going to add new
    # attributes from within the loop
    for name in vars(cli_args).copy():
        value = getattr(cli_args, name)

        # Solving wrapped config values
        if isinstance(value, WrappedConfigValue):
            setattr(cli_args, name, value.resolve(config))

        # Resolving dummy csv input files
        if isinstance(value, DummyCSVInput):
            value = value.resolve(cli_args)
            setattr(cli_args, name, value)

        # Opening output files
        if isinstance(value, OutputOpener):
            value = value.open(cli_args, resume=getattr(cli_args, "resume", False))
            setattr(cli_args, name, value)

        # Finding buffers to close eventually
        if (
            isinstance(value, TextIOBase)
            and value is not sys.stdin
            and value is not sys.stdout
            and value is not sys.stderr
        ) or isinstance(value, Resumer):
            to_close.append(value)

    return to_close


class VariadicInputDefinition(TypedDict):
    dummy_column: str
    item_label: NotRequired[str]
    item_label_plural: NotRequired[str]
    column_help: NotRequired[str]
    input_help: NotRequired[str]
    no_help: NotRequired[bool]
    optional: NotRequired[bool]
    metavar: NotRequired[str]


def resolve_typical_arguments(
    args,
    no_output=False,
    resumer=None,
    resumer_kwargs=None,
    select: bool = False,
    total: bool = False,
    variadic_input: Optional[VariadicInputDefinition] = None,
):
    args = [] if args is None else args.copy()
    epilog_addendum = None

    output_argument = {"flags": ["-o", "--output"], "action": OutputAction}

    if variadic_input is not None:
        variadic_input = variadic_input.copy()

        if "item_label" not in variadic_input:
            variadic_input["item_label"] = variadic_input["dummy_column"]

        if "item_label_plural" not in variadic_input:
            variadic_input["item_label_plural"] = variadic_input["item_label"] + "s"

        if "column_help" not in variadic_input:
            variadic_input["column_help"] = (
                "Single %(singular)s to process or name of the CSV column containing %(plural)s when using -i/--input."
                % {
                    "singular": variadic_input["item_label"],
                    "plural": variadic_input["item_label_plural"],
                }
            )

            if variadic_input.get("optional", False):
                variadic_input["column_help"] += ' Defaults to "{}".'.format(
                    variadic_input["dummy_column"]
                )

        if "input_help" not in variadic_input:
            variadic_input["input_help"] = (
                "CSV file containing all the %s you want to process. Will consider `-` as stdin."
                % variadic_input["item_label_plural"]
            )

        column_arg = {
            "name": "column",
            "metavar": variadic_input.get(
                "metavar",
                "{item}_or_{item}_column".format(item=variadic_input["dummy_column"]),
            ),
            "help": variadic_input["column_help"],
        }

        if variadic_input.get("optional", False):
            column_arg["nargs"] = "?"

        args.append(column_arg)

        input_argument = {
            "flags": ["-i", "--input"],
            "help": variadic_input["input_help"],
            "action": InputAction,
            "dummy_csv_column": variadic_input["dummy_column"],
        }

        args.append(input_argument)

        args.append(
            {
                "flags": ["--explode"],
                "help": "Use to indicate the character used to separate multiple values in a single CSV cell. Defaults to none, i.e. CSV cells having a single values, which is usually the case.",
            }
        )

        if not variadic_input.get("no_help", False):
            epilog_addendum = """
        how to use the command with a CSV file?

        > A lot of minet commands, including this one, can both be
        > given a single value to process or a bunch of them if
        > given the column of a CSV file passed to -i/--input instead.

        . Here is how to use a command with a single value:
            $ minet cmd "value"

        . Here is how to use a command with a csv file:
            $ minet cmd column_name -i file.csv

        . Here is how to read CSV file from stdin using `-`:
            $ xsv search -s col . | minet cmd column_name -i -

        . Here is how to indicate that the CSV column may contain multiple
          values separated by a special character:
            $ minet cmd column_name -i file.csv --explode "|"

        . This also works with single values:
            $ minet cmd "value1,value2" --explode ","
        """

    if select or variadic_input is not None:

        # TODO: actually one can use xsv mini dsl here
        args.append(
            {
                "flags": ["-s", "--select"],
                "help": "Columns of input CSV file to include in the output (separated by `,`).",
            },
        )

    if total or variadic_input is not None:
        args.append(
            {
                "flag": "--total",
                "help": "Total number of items to process. Might be necessary when you want to display a finite progress indicator for large files given as input to the command.",
                "type": int,
            }
        )

    if resumer is not None:
        args.append(
            {
                "flag": "--resume",
                "help": "Whether to resume from an aborted collection. Need -o to be set.",
                "action": "store_true",
            },
        )

        output_argument["resumer"] = resumer

        if resumer_kwargs is not None:
            output_argument["resumer_kwargs"] = resumer_kwargs

    if not no_output:
        args.append(output_argument)

    return args, epilog_addendum


def command(
    name: str,
    package=None,
    title: Optional[str] = None,
    aliases=None,
    description=None,
    epilog=None,
    common_arguments=None,
    arguments=None,
    subcommands=None,
    resolve=None,
    resumer=None,
    resumer_kwargs=None,
    no_output=False,
    select=False,
    total=False,
    variadic_input: Optional[VariadicInputDefinition] = None,
    **kwargs,
):

    if resumer is not None and not issubclass(resumer, Resumer):
        raise TypeError("given resumer is not a Resumer class")

    if arguments is not None and subcommands is not None:
        raise TypeError(
            "command cannot have subcommands and be executable on its own (avoid giving arguments AND subcommands)"
        )

    data = {"name": name, "title": title}

    if package is not None:
        data["package"] = package

    if aliases is not None:
        data["aliases"] = aliases

    if description is not None:
        data["description"] = description

    if epilog is not None:
        data["epilog"] = epilog

    if subcommands is not None:
        data["subparsers"] = {
            "help": "Subcommand to use.",
            "title": "subcommands",
            "dest": "subcommand",
            "commands": {s["name"]: s for s in subcommands},
        }

        if common_arguments is not None:
            data["subparsers"]["common_arguments"] = common_arguments

    elif arguments is not None:
        data["arguments"], epilog_addendum = resolve_typical_arguments(
            arguments,
            no_output=no_output,
            resumer=resumer,
            resumer_kwargs=resumer_kwargs,
            select=select,
            total=total,
            variadic_input=variadic_input,
        )

        if epilog_addendum is not None:
            if not "epilog" in data:
                data["epilog"] = epilog_addendum
            else:
                data["epilog"] += "\n\n" + epilog_addendum

    if resolve is not None:
        data["resolve"] = resolve

    data.update(kwargs)

    return data


def subcommand(
    name,
    package,
    title,
    description=None,
    epilog=None,
    arguments=[],
    resolve=None,
    resumer=None,
    resumer_kwargs=None,
    no_output=False,
    select=False,
    total=False,
    variadic_input: Optional[VariadicInputDefinition] = None,
    **kwargs,
):
    data = {"name": name, "title": title, "package": package, "arguments": arguments}

    if description is not None:
        data["description"] = description

    if epilog is not None:
        data["epilog"] = epilog

    data["arguments"], epilog_addendum = resolve_typical_arguments(
        arguments,
        no_output=no_output,
        resumer=resumer,
        resumer_kwargs=resumer_kwargs,
        select=select,
        total=total,
        variadic_input=variadic_input,
    )

    if epilog_addendum is not None:
        if not "epilog" in data:
            data["epilog"] = epilog_addendum
        else:
            data["epilog"] += "\n\n" + epilog_addendum

    if resolve is not None:
        data["resolve"] = resolve

    data.update(kwargs)

    return data


def template_readme(tpl, commands):
    _, subparser_index = build_parser("", "", commands)

    def replacer(match):
        keys = match.group(1).split("/")

        target = get_subparser(subparser_index, keys)

        if target is None:
            raise TypeError('missing key "%s"' % "/".join(keys))

        return (
            dedent(
                """
                    ```
                    %s
                    ```
                """
            )
            % target.format_help().strip()
        ).strip()

    return re.sub(TEMPLATE_RE, replacer, tpl)



from minet.cli.utils import with_fatal_errors

from minet.mediacloud.exceptions import MediacloudServerError


def fatal_errors_hook(e):
    if isinstance(e, MediacloudServerError):
        return ["Aborted due to a mediacloud server error:", e.server_error]


def with_mediacloud_fatal_errors(fn):
    return with_fatal_errors(fatal_errors_hook)(fn)



# =============================================================================
# Minet Scraper Utils
# =============================================================================
#
# Miscellaneous helper functions used throughout the minet.scrape package.
#
from typing import Optional
from minet.types import AnyScrapableTarget

import warnings
from bs4 import BeautifulSoup, XMLParsedAsHTMLWarning, SoupStrainer
from functools import partial

from minet.scrape.constants import SELECT_ALIASES, ITERATOR_ALIASES


def get_aliases(aliases, target, with_key=False):
    for alias in aliases:
        if alias in target:
            if with_key:
                return alias, target[alias]

            return target[alias]

    if with_key:
        return None, None

    return None


get_sel = partial(get_aliases, SELECT_ALIASES)
get_iterator = partial(get_aliases, ITERATOR_ALIASES)


def BeautifulSoupWithoutXHTMLWarnings(html, engine):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=XMLParsedAsHTMLWarning)
        return BeautifulSoup(html, engine)


def ensure_soup(
    html_or_soup: AnyScrapableTarget,
    engine: str = "lxml",
    strainer: Optional[SoupStrainer] = None,
) -> BeautifulSoup:
    is_already_soup = isinstance(html_or_soup, BeautifulSoup)

    if not is_already_soup:
        return BeautifulSoup(html_or_soup, engine, parse_only=strainer)

    return html_or_soup



# =============================================================================
# Minet Telegram Channel-Infos CLI Action
# =============================================================================
#
# Action retrieving the information of a Telegram channel.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.telegram import TelegramScraper
from minet.telegram.constants import TELEGRAM_INFOS_CSV_HEADERS
from minet.telegram.exceptions import TelegramInvalidTargetError


@with_enricher_and_loading_bar(
    headers=TELEGRAM_INFOS_CSV_HEADERS, title="Retrieving info", unit="channels"
)
def action(cli_args, enricher, loading_bar):
    scraper = TelegramScraper(throttle=cli_args.throttle)

    for i, row, channel in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step():
            try:
                infos = scraper.channel_infos(channel)
                enricher.writerow(row, infos)
            except TelegramInvalidTargetError:
                loading_bar.print(
                    "%s (line %i) is not a telegram channel or url, or is not accessible."
                    % (channel, i)
                )



# =============================================================================
# Minet Instagram Formatters
# =============================================================================
#
# Various data formatters for Instagram data.
#
from casanova import namedrecord
from ebbe import getpath

from minet.instagram.constants import (
    INSTAGRAM_COMMENT_CSV_HEADERS,
    INSTAGRAM_HASHTAG_POST_CSV_HEADERS,
    INSTAGRAM_POST_CSV_HEADERS,
    INSTAGRAM_MEDIA_TYPE,
    INSTAGRAM_USER_CSV_HEADERS,
    INSTAGRAM_USER_INFO_CSV_HEADERS,
)
from minet.instagram.utils import (
    extract_hashtags,
    extract_handles,
    short_code_to_url,
)
from minet.utils import timestamp_to_isoformat

INSTAGRAM_PATH_TO_MEDIA_IMAGE = ["image_versions2", "candidates", 0, "url"]
INSTAGRAM_PATH_TO_MEDIA_VIDEO = ["video_versions", 0, "url"]

InstagramComment = namedrecord(
    "InstagramComment",
    INSTAGRAM_COMMENT_CSV_HEADERS,
    boolean=[
        "has_translation",
        "is_liked_by_media_owner",
        "user_is_private",
        "user_is_mentionable",
        "user_is_verified",
    ],
    plural=[
        "hashtags",
        "mentioned_names"
    ],
)

InstagramHashtagPost = namedrecord(
    "InstagramHashtagPost",
    INSTAGRAM_HASHTAG_POST_CSV_HEADERS,
    boolean=[
        "comments_disabled",
    ],
    plural=["hashtags", "mentioned_names"],
)

InstagramPost = namedrecord(
    "InstagramPost",
    INSTAGRAM_POST_CSV_HEADERS,
    boolean=["like_and_view_counts_disabled", "is_verified"],
    plural=[
        "medias_type",
        "medias_url",
        "usertags_medias",
        "mentioned_names",
        "hashtags",
    ],
)

InstagramUser = namedrecord(
    "InstagramUser",
    INSTAGRAM_USER_CSV_HEADERS,
    boolean=[
        "is_private",
        "is_verified",
        "has_anonymous_profile_picture",
        "has_highlight_reels",
        "is_favorite",
    ],
)

InstagramUserInfo = namedrecord(
    "InstagramUserInfo",
    INSTAGRAM_USER_INFO_CSV_HEADERS,
    boolean=[
        "is_private",
        "is_verified",
        "hide_like_and_view_counts",
        "has_ar_effects",
        "has_clips",
        "has_guides",
        "has_channel",
        "is_business_account",
        "is_professional_account",
        "is_supervision_enabled",
        "is_guardian_of_viewer",
        "is_supervised_by_viewer",
        "is_supervised_user",
        "is_joined_recently",
    ],
    plural=[
        "bio_links_title",
        "bio_links_url",
        "biography_with_username",
        "biography_with_hashtag",
        "pronouns",
    ],
)


def get_usertags(item):
    users = getpath(item, ["usertags", "in"])
    usertags = set()
    if not users:
        return usertags
    for user in users:
        username = getpath(user, ["user", "username"])
        usertags.add(username)
    return usertags


def format_comment(comment):
    text = comment.get("text")

    hashtags = []
    mentioned_names = []

    if comment.get("text"):
        hashtags = extract_hashtags(text)
        mentioned_names = extract_handles(text)

    row = InstagramComment(
        comment.get("pk"),
        comment.get("comment_index"),
        comment.get("parent_comment_id"),
        comment.get("child_comment_index"),
        comment.get("text"),
        hashtags,
        mentioned_names,
        comment.get("has_translation", False),
        comment.get("comment_like_count"),
        comment.get("is_liked_by_media_owner"),
        comment.get("child_comment_count"),
        comment.get("created_at_utc"),
        timestamp_to_isoformat(comment.get("created_at_utc")),
        getpath(comment, ["user", "username"]),
        comment.get("user_id"),
        getpath(comment, ["user", "full_name"]),
        getpath(comment, ["user", "is_private"]),
        getpath(comment, ["user", "is_mentionable"]),
        getpath(comment, ["user", "is_verified"]),
        getpath(comment, ["user", "profile_pic_id"]),
        getpath(comment, ["user", "profile_pic_url"]),
        getpath(comment, ["user", "fbid_v2"]),
    )

    return row


def format_hashtag_post(item):

    text = getpath(item, ["edge_media_to_caption", "edges", 0, "node", "text"])

    hashtags = []
    mentioned_names = []

    if text:
        hashtags = extract_hashtags(text)
        mentioned_names = extract_handles(text)

    row = InstagramHashtagPost(
        item["id"],
        item["__typename"],
        item["shortcode"],
        short_code_to_url(item["shortcode"]),
        text,
        hashtags,
        mentioned_names,
        item["edge_liked_by"]["count"],
        item["comments_disabled"],
        item["edge_media_to_comment"]["count"],
        item["edge_media_preview_like"]["count"],
        item.get("video_view_count"),
        item["owner"]["id"],
        item["taken_at_timestamp"],
        timestamp_to_isoformat(item["taken_at_timestamp"]),
        item["accessibility_caption"],
        short_code_to_url(item["shortcode"]) + "media/?size=l",
    )

    return row


def format_post(item):

    media_type = INSTAGRAM_MEDIA_TYPE.get(item["media_type"])
    if media_type is None:
        media_type = item["media_type"]

    text = getpath(item, ["caption", "text"])

    hashtags = []
    mentioned_names = []

    if text:
        hashtags = extract_hashtags(text)
        mentioned_names = extract_handles(text)

    medias_url = []
    medias_type = []
    usertags_medias = set()

    if media_type == "GraphSidecar":
        carousel = getpath(item, ["carousel_media"])
        for media in carousel:
            medias_type.append(INSTAGRAM_MEDIA_TYPE.get(getpath(media, ["media_type"])))

            if medias_type[-1] == "GraphImage":
                medias_url.append(getpath(media, INSTAGRAM_PATH_TO_MEDIA_IMAGE))
                item_usertags = get_usertags(media)
                usertags_medias = usertags_medias.union(item_usertags)

            elif medias_type[-1] == "GraphVideo":
                medias_url.append(getpath(media, INSTAGRAM_PATH_TO_MEDIA_VIDEO))
                item_usertags = get_usertags(media)
                usertags_medias = usertags_medias.union(item_usertags)

    elif media_type == "GraphImage":
        medias_type.append("GraphImage")
        medias_url.append(getpath(item, INSTAGRAM_PATH_TO_MEDIA_IMAGE))
        item_usertags = get_usertags(item)
        usertags_medias = usertags_medias.union(item_usertags)

    elif media_type == "GraphVideo":
        medias_type.append("GraphVideo")
        medias_url.append(getpath(item, INSTAGRAM_PATH_TO_MEDIA_VIDEO))
        item_usertags = get_usertags(item)
        usertags_medias = usertags_medias.union(item_usertags)

    row = InstagramPost(
        getpath(item, ["user", "username"]),
        getpath(item, ["user", "full_name"]),
        getpath(item, ["user", "is_verified"]),
        item["pk"],
        media_type,
        item["code"],
        short_code_to_url(item["code"]),
        text,
        hashtags,
        mentioned_names,
        item["like_and_view_counts_disabled"],
        item["like_count"],
        item["comment_count"],
        short_code_to_url(item["code"]) + "media/?size=l",
        medias_type,
        medias_url,
        usertags_medias,
        item["taken_at"],
        timestamp_to_isoformat(item["taken_at"]),
    )

    return row


def format_user(item):

    row = InstagramUser(
        item.get("pk"),
        item.get("username"),
        item.get("full_name"),
        item.get("is_private"),
        item.get("profile_pic_url"),
        item.get("profile_pic_id"),
        item.get("is_verified"),
        item.get("has_anonymous_profile_picture"),
        item.get("has_highlight_reels"),
        item.get("account_badges"),
        item.get("similar_user_id"),
        item.get("latest_reel_media"),
        item.get("is_favorite"),
    )

    return row


def format_user_info(user):

    bio_links_title = []
    bio_links_url = []
    biography_with_username = []
    biography_with_hashtag = []

    bio_links = user.get("bio_links")
    if bio_links:
        for link in bio_links:
            bio_links_title.append(link.get("title"))
            bio_links_url.append(link.get("url"))

    biography_with_entities = getpath(user, ["biography_with_entities", "entities"])

    if biography_with_entities:

        for entity in biography_with_entities:
            username = getpath(entity, ["user", "username"])
            if username:
                biography_with_username.append(username)

            hashtag = getpath(entity, ["user", "hashtag"])
            if hashtag:
                biography_with_hashtag.append(hashtag)

    pronouns = user.get("pronouns")

    row = InstagramUserInfo(
        user.get("username"),
        user.get("id"),
        user.get("full_name"),
        user.get("profile_pic_url_hd"),
        user.get("is_private"),
        user.get("is_verified"),
        user.get("biography"),
        bio_links_title,
        bio_links_url,
        biography_with_username,
        biography_with_hashtag,
        getpath(user, ["edge_followed_by", "count"]),
        getpath(user, ["edge_follow", "count"]),
        user.get("hide_like_and_view_counts"),
        pronouns,
        user.get("has_ar_effects"),
        user.get("has_clips"),
        user.get("has_guides"),
        user.get("has_channel"),
        user.get("is_business_account"),
        user.get("is_professional_account"),
        user.get("is_supervision_enabled"),
        user.get("is_guardian_of_viewer"),
        user.get("is_supervised_by_viewer"),
        user.get("is_supervised_user"),
        user.get("guardian_id"),
        user.get("is_joined_recently"),
        user.get("business_category_name"),
        user.get("category_name"),
        user.get("connected_fb_page"),
    )

    return row



from typing import Iterable, Iterator, TypeVar, Callable, Tuple, List, Optional

from urllib.parse import quote, unquote

from minet.fetch import FetchThreadPoolExecutor

from minet.wikipedia.exceptions import (
    WikimediaRESTAPIThrottledError,
    WikimediaRESTAPIServerError,
)
from minet.wikipedia.types import Granularity, Agent, Access, WikipediaPageViewsItem

BASE_URL = "https://wikimedia.org/api/rest_v1"
DEFAULT_THREADS = 10

ItemType = TypeVar("ItemType")


# NOTE: this is not a sure shot, but good enough for wikipedia names
def ensure_is_quoted(string: str) -> str:
    if string == unquote(string):
        return string

    return quote(string)


def fix_date(string: str) -> str:
    if len(string) == 8:
        return string + "00"

    return string


def build_pageviews_url(
    name: str,
    lang: str,
    start_date: str,
    end_date: str,
    access: Access = "all-access",
    agent: Agent = "all-agents",
    granularity: Granularity = "monthly",
) -> str:
    return "{base_url}/metrics/pageviews/per-article/{lang}.wikipedia/{access}/{agent}/{name}/{granularity}/{start_date}/{end_date}".format(
        base_url=BASE_URL,
        name=ensure_is_quoted(name),
        lang=lang,
        granularity=granularity,
        access=access,
        agent=agent,
        start_date=start_date,
        end_date=end_date,
    )


class WikimediaRestAPIClient(object):
    def pageviews(
        self,
        pages: Iterable[ItemType],
        start_date: str,
        end_date: str,
        access: Access = "all-access",
        agent: Agent = "all-agents",
        granularity: Granularity = "monthly",
        key: Optional[Callable[[ItemType], Tuple[str, str]]] = None,
        lang: Optional[str] = None,
        threads: int = DEFAULT_THREADS,
    ) -> Iterator[Tuple[ItemType, List[WikipediaPageViewsItem]]]:
        start_date = fix_date(start_date)
        end_date = fix_date(end_date)

        def api_url_from_item(page) -> str:
            page = page if not callable(key) else key(page)

            if lang is not None:
                if not isinstance(page, tuple):
                    page = (lang, page)
                else:
                    page = (lang, page[1])

            url = build_pageviews_url(
                lang=page[0],
                name=page[1],
                start_date=start_date,
                end_date=end_date,
                access=access,
                granularity=granularity,
                agent=agent,
            )

            return url

        with FetchThreadPoolExecutor(max_workers=threads, retry=True) as executor:
            for result in executor.imap_unordered(
                pages,
                key=api_url_from_item,
                domain_parallelism=threads,
                throttle=0,
            ):
                if result.error:
                    raise result.error

                response = result.response

                assert response is not None

                if response.status == 429:
                    raise WikimediaRESTAPIThrottledError

                if response.status >= 500:
                    raise WikimediaRESTAPIServerError

                if response.status == 404:
                    yield result.item, []
                    continue

                pageviews = []

                for item in response.json()["items"]:
                    pageviews.append(
                        WikipediaPageViewsItem(
                            project=item["project"],
                            article=item["article"],
                            granularity=item["granularity"],
                            timestamp=int(item["timestamp"]),
                            access=item["access"],
                            agent=item["agent"],
                            views=item["views"],
                        )
                    )

                yield result.item, pageviews



# =============================================================================
# Minet Youtube CLI Action
# =============================================================================
#
# Logic of the `yt` action.
#
from minet.cli.argparse import command, subcommand, ConfigAction, SplitterType

# TODO: this is a lazyloading issue
from minet.youtube.constants import (
    YOUTUBE_API_DEFAULT_SEARCH_ORDER,
    YOUTUBE_API_SEARCH_ORDERS,
)

KEY_ARGUMENT = {
    "flags": ["-k", "--key"],
    "help": "YouTube API Data dashboard API key. Can be used more than once.",
    "rc_key": ["youtube", "key"],
    "action": ConfigAction,
    "plural": True,
    "required": True,
}


def youtube_api_subcommand(*args, arguments=[], **kwargs):
    return subcommand(*args, arguments=arguments + [KEY_ARGUMENT], **kwargs)


YOUTUBE_CAPTIONS_SUBCOMMAND = subcommand(
    "captions",
    "minet.cli.youtube.captions",
    title="Youtube captions",
    description="Retrieve captions for the given YouTube videos.",
    epilog="""
        Examples:

        . Fetching captions for a list of videos:
            $ minet yt captions video_id videos.csv > captions.csv

        . Fetching French captions with a fallback to English:
            $ minet yt captions video_id videos.csv --lang fr,en > captions.csv
    """,
    variadic_input={
        "dummy_column": "video",
        "item_label": "video url or id",
        "item_label_plural": "video urls or ids",
    },
    arguments=[
        {
            "flag": "--lang",
            "help": 'Language (ISO code like "en") of captions to retrieve. You can specify several languages by preferred order separated by commas. Defaults to "en".',
            "default": ["en"],
            "type": SplitterType(),
        }
    ],
)

YOUTUBE_CHANNEL_VIDEOS_SUBCOMMAND = youtube_api_subcommand(
    "channel-videos",
    "minet.cli.youtube.channel_videos",
    title="Youtube channel videos",
    description="""
        Retrieve metadata about all Youtube videos from one or many channel(s) using the API.

        Under the hood, this command extract the channel id from the given url or scrape the
        website to find it if necessary. Then the command uses the API to retrieve
        information about videos stored in the main playlist of the channel
        supposed to contain all the channel's videos.
    """,
    epilog="""
        example:

        . Fetching all the videos from a channel based on the channel's id or url:
            $ minet youtube channel-videos https://www.youtube.com/c/LinksOff -k my-api-key > linksoff_videos.csv
            $ minet youtube channel-videos https://www.youtube.com/channel/UCqnbDFdCpuN8CMEg0VuEBqA -k my-api-key > thenewyorktimes_videos.csv
            $ minet youtube channel-videos UCprclkVrNPls7PR-nHhf1Ow -k my-api-key > tonyheller_videos.csv

        . Fetching multiple channels' videos:
            $ minet youtube channel-videos channel_id channels_id.csv -k my-api-key > channels_videos.csv
            $ minet youtube channel-videos channel_url channels_url.csv -k my-api-key > channels_videos.csv
    """,
    variadic_input={"dummy_column": "channel"},
)

YOUTUBE_CHANNELS_SUBCOMMAND = youtube_api_subcommand(
    "channels",
    "minet.cli.youtube.channels",
    title="Youtube Channels Command",
    description="""
        Retrieve metadata about Youtube channel from one or many name(s) using the API.

        Under the hood, this command extract the channel id from the given url or scrape the
        website to find it if necessary. Then the command uses the API to retrieve
        information about the channel.
    """,
    epilog="""
        example:

        . Fetching metadata from a channel based on the channel's id or url:
            $ minet youtube channels https://www.youtube.com/c/LinksOff -k my-api-key > linksoff_meta.csv
            $ minet youtube channels https://www.youtube.com/channel/UCqnbDFdCpuN8CMEg0VuEBqA -k my-api-key > thenewyorktimes_meta.csv
            $ minet youtube channels UCprclkVrNPls7PR-nHhf1Ow -k my-api-key > tonyheller_meta.csv

        . Fetching multiple channels' metadata:
            $ minet youtube channels channel_id channels_id.csv -k my-api-key > channels.csv
            $ minet youtube channels channel_url channels_url.csv -k my-api-key > channels.csv
    """,
    variadic_input={"dummy_column": "channel"},
)

YOUTUBE_COMMENTS_SUBCOMMAND = youtube_api_subcommand(
    "comments",
    "minet.cli.youtube.comments",
    title="Youtube comments",
    description="Retrieve metadata about Youtube comments using the API.",
    epilog="""
        example:

        . Fetching a video's comments:
            $ minet yt comments https://www.youtube.com/watch?v=7JTb2vf1OQQ -k my-api-key > comments.csv
    """,
    variadic_input={"dummy_column": "video"},
)

YOUTUBE_SEARCH_SUBCOMMAND = youtube_api_subcommand(
    "search",
    "minet.cli.youtube.search",
    title="Youtube search",
    description="""
        Search videos using the YouTube API.

        Note that, even if undocumented, the API will never return
        more than approx. 500 videos for a given query.
    """,
    epilog="""
        example:

        . Searching videos about birds:
            $ minet youtube search bird -k my-api-key > bird_videos.csv
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of videos to retrieve per query.",
            "type": int,
        },
        {
            "flags": ["--order"],
            "help": "Order in which videos are retrieved. The default one is relevance.",
            "default": YOUTUBE_API_DEFAULT_SEARCH_ORDER,
            "choices": YOUTUBE_API_SEARCH_ORDERS,
        },
    ],
)

YOUTUBE_VIDEOS_SUBCOMMAND = youtube_api_subcommand(
    "videos",
    "minet.cli.youtube.videos",
    title="Youtube videos",
    description="Retrieve metadata about Youtube videos using the API.",
    variadic_input={"dummy_column": "video"},
)

YOUTUBE_COMMAND = command(
    "youtube",
    "minet.cli.youtube",
    "Minet YouTube Command",
    aliases=["yt"],
    description="""
        Gather data from Youtube.
    """,
    subcommands=[
        YOUTUBE_CAPTIONS_SUBCOMMAND,
        YOUTUBE_CHANNEL_VIDEOS_SUBCOMMAND,
        YOUTUBE_CHANNELS_SUBCOMMAND,
        YOUTUBE_COMMENTS_SUBCOMMAND,
        YOUTUBE_SEARCH_SUBCOMMAND,
        YOUTUBE_VIDEOS_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Url Extract CLI Action
# =============================================================================
#
# Logic of the `url-extract` action.
#
from ural import urls_from_text, urls_from_html
from urllib.parse import urljoin

from minet.cli.utils import with_enricher_and_loading_bar

REPORT_HEADERS = ["extracted_url"]

EXTRACTORS = {"html": urls_from_html, "text": urls_from_text}


@with_enricher_and_loading_bar(
    headers=REPORT_HEADERS, title="Extracting urls", unit="docs"
)
def action(cli_args, enricher, loading_bar):
    extract = EXTRACTORS[getattr(cli_args, "from")]

    for row, content in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            content = content.strip()

            if not content:
                continue

            for url in extract(content):
                if cli_args.base_url is not None:
                    url = urljoin(cli_args.base_url, url)

                enricher.writerow(row, [url])



# =============================================================================
# Minet Mediacloud Search CLI Action
# =============================================================================
#
# Logic of the `mc search` action.
#
import casanova

from minet.cli.loading_bar import LoadingBar
from minet.cli.mediacloud.utils import with_mediacloud_fatal_errors
from minet.mediacloud import MediacloudAPIClient
from minet.mediacloud.constants import MEDIACLOUD_STORIES_CSV_HEADER


@with_mediacloud_fatal_errors
def action(cli_args):
    writer = casanova.writer(cli_args.output, fieldnames=MEDIACLOUD_STORIES_CSV_HEADER)

    client = MediacloudAPIClient(cli_args.token)

    kwargs = {
        "collections": cli_args.collections,
        "medias": cli_args.medias,
        "publish_day": cli_args.publish_day,
        "publish_month": cli_args.publish_month,
        "publish_year": cli_args.publish_year,
        "filter_query": cli_args.filter_query,
    }

    with LoadingBar(title="Searching", unit="stories") as loading_bar:
        if not cli_args.skip_count:
            count = client.count(cli_args.query, **kwargs)
            loading_bar.set_total(count)

        iterator = client.search(cli_args.query, **kwargs)

        for story in iterator:
            writer.writerow(story.as_csv_row())
            loading_bar.update()



# =============================================================================
# Minet Hyphe API Client
# =============================================================================
#
# Client for the Hyphe API.
#
from minet.web import create_pool_manager, request_jsonrpc
from minet.hyphe.constants import WEBENTITY_STATUSES, DEFAULT_PAGINATION_COUNT
from minet.hyphe.exceptions import (
    HypheCouldNotStartCorpusError,
    HypheJSONRPCError,
    HypheRequestFailError,
    HypheCorpusAuthenticationError,
)


class HypheAPIClient(object):
    def __init__(self, endpoint):
        if not endpoint.endswith("/"):
            endpoint += "/"

        self.endpoint = endpoint
        self.pool_manager = create_pool_manager()

    def call(self, method, *args, **kwargs):
        response = request_jsonrpc(
            self.endpoint, method, pool_manager=self.pool_manager, *args, **kwargs
        )

        result = response.json()

        if "fault" in result:
            return HypheJSONRPCError(result), None

        if (
            isinstance(result, list)
            and len(result) > 0
            and isinstance(result[0], dict)
            and result[0].get("code") == "fail"
        ):
            msg = result[0].get("message", "")

            if "password" in msg:
                raise HypheCorpusAuthenticationError

            raise HypheRequestFailError(msg)

        return result[0]

    def corpus(self, corpus, password=None):
        return HypheAPIClientCorpus(self, corpus, password)


class HypheAPIClientCorpus(object):
    def __init__(self, client, corpus, password=None):
        self.client = client
        self.corpus = corpus
        self.password = password

    def call(self, method, *args, **kwargs):
        kwargs["corpus"] = self.corpus

        return self.client.call(method, *args, **kwargs)

    def ping_until_ready(self, max_attempts=10):
        for _ in range(max_attempts):
            result = self.call("ping", timeout=5)

            if result["result"] == "pong":
                return True

        raise HypheCouldNotStartCorpusError

    def ensure_is_started(self):
        result = self.call("start_corpus", password=self.password)

        # Corpus is already started
        if result["result"]["status"] == "ready":
            return True

        # Pinging until ready
        return self.ping_until_ready()

    def status(self):
        stats = self.call("get_status")

        return stats

    def count(self, statuses=None):
        if not statuses:
            statuses = WEBENTITY_STATUSES

        stats = self.status()

        webentity_counts = stats["result"]["corpus"]["traph"]["webentities"]

        total = 0

        for status in statuses:
            total += webentity_counts[status]

        info = {
            "webentities": total,
            "pages": stats["result"]["corpus"]["crawler"]["pages_found"],
        }

        return info

    def webentities(self, statuses=None):
        if not statuses:
            statuses = WEBENTITY_STATUSES

        for status in statuses:
            token = None
            next_page = None

            while True:
                if token is None:
                    result = self.call(
                        "store.get_webentities_by_status",
                        status=status,
                        count=DEFAULT_PAGINATION_COUNT,
                    )
                else:
                    result = self.call(
                        "store.get_webentities_page",
                        pagination_token=token,
                        n_page=next_page,
                    )

                result = result["result"]

                for webentity in result["webentities"]:
                    yield webentity

                if "next_page" in result and result["next_page"]:
                    token = result["token"]
                    next_page = result["next_page"]
                else:
                    break

    def webentity_pages(self, webentity_id, include_body=False):
        token = None

        while True:
            result = self.call(
                "store.paginate_webentity_pages",
                webentity_id=webentity_id,
                count=DEFAULT_PAGINATION_COUNT,
                pagination_token=token,
                include_page_metas=True,
                include_page_body=include_body,
            )

            result = result["result"]

            for page in result["pages"]:
                yield page

            if "token" in result and result["token"]:
                token = result["token"]
            else:
                break



# =============================================================================
# Minet Twitter Retweeters CLI Action
# =============================================================================
#
# Logic of the `tw retweeters` action.
#
import re
from ebbe import getpath
from twitter import TwitterHTTPError
from twitwi import normalize_user, format_user_as_csv_row
from twitwi.constants import USER_PARAMS, USER_FIELDS
from ural.twitter import parse_twitter_url, TwitterTweet, TwitterUser, TwitterList

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.twitter.utils import with_twitter_client

ID_RE = re.compile(r"^[0-9]+$")

ITEMS_PER_PAGE = 100


@with_enricher_and_loading_bar(
    headers=USER_FIELDS,
    title="Retrieving retweeters",
    unit="tweets",
    nested=True,
    sub_unit="retweeters",
)
@with_twitter_client(api_version="2")
def action(cli_args, client, enricher, loading_bar):
    for row, tweet in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(tweet):
            tweet_id = tweet

            tweet_parsed = parse_twitter_url(tweet)
            if isinstance(tweet_parsed, TwitterTweet):
                tweet_id = tweet_parsed.id
            elif isinstance(tweet_parsed, (TwitterUser, TwitterList)) or (
                not tweet_parsed and not ID_RE.match(tweet)
            ):
                loading_bar.print("%s is not a tweet id or url." % tweet)
                continue

            kwargs = {"max_results": ITEMS_PER_PAGE, "params": USER_PARAMS}

            while True:
                try:
                    result = client.call(["tweets", tweet_id, "retweeted_by"], **kwargs)
                except TwitterHTTPError as e:
                    if e.e.code == 404:
                        enricher.writerow(row)
                    else:
                        raise e

                    break

                if result.get("errors"):
                    loading_bar.print(getpath(result, ["errors", 0, "detail"]))
                    break

                if "data" not in result or result["meta"]["result_count"] == 0:
                    break

                for user in result["data"]:
                    user = normalize_user(user, v2=True)
                    user_row = format_user_as_csv_row(user)
                    enricher.writerow(row, user_row)
                    loading_bar.nested_advance()

                if "next_token" in result["meta"]:
                    kwargs["pagination_token"] = result["meta"]["next_token"]
                else:
                    break
